# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging  # noqa: F401
from collections import abc  # noqa: F401
from typing import Any, List, Optional, Sequence  # noqa: F401

from opentelemetry.proto.collector.trace.v1.trace_service_pb2 import (  # noqa: F401
    ExportTraceServiceRequest as PB2ExportTraceServiceRequest,
)
from opentelemetry.proto.common.v1.common_pb2 import (  # noqa: F401
    AnyValue as PB2AnyValue,
)
from opentelemetry.proto.common.v1.common_pb2 import (  # noqa: F401
    ArrayValue as PB2ArrayValue,
)
from opentelemetry.proto.common.v1.common_pb2 import (  # noqa: F401
    InstrumentationScope as PB2InstrumentationScope,
)
from opentelemetry.proto.common.v1.common_pb2 import (  # noqa: F401
    KeyValue as PB2KeyValue,
)
from opentelemetry.proto.resource.v1.resource_pb2 import (  # noqa: F401
    Resource as PB2Resource,
)
from opentelemetry.proto.trace.v1.trace_pb2 import (  # noqa: F401
    ScopeSpans as PB2ScopeSpans,
)
from opentelemetry.proto.trace.v1.trace_pb2 import (  # noqa: F401
    ResourceSpans as PB2ResourceSpans,
)
from opentelemetry.proto.trace.v1.trace_pb2 import (  # noqa: F401
    Span as PB2SPan,
)
from opentelemetry.proto.trace.v1.trace_pb2 import (  # noqa: F401
    Status as PB2Status,
)
from opentelemetry.sdk.trace import Event  # noqa: F401
from opentelemetry.sdk.util.instrumentation import (  # noqa: F401
    InstrumentationScope,
)
from opentelemetry.sdk.trace import Resource  # noqa: F401
from opentelemetry.sdk.trace import Span as SDKSpan  # noqa: F401
from opentelemetry.trace import Link  # noqa: F401
from opentelemetry.trace import SpanKind  # noqa: F401
from opentelemetry.trace.span import (  # noqa: F401
    SpanContext,
    TraceState,
    Status,
)
from opentelemetry.util.types import Attributes  # noqa: F401



import json
import logging

from datetime import (
    datetime,
    timedelta,
)
from threading import Lock
from typing import Dict

import pyscripts
from bustoken import TK_SLCTRL_PMGR_GET_PRODUCTS
from msgbus import (
    FM_PARAM,
    TK_SYS_ACK,
    MBEasyContext,
)
from sysactions import (
    action,
    StopAction,
)

from model.customexception import (
    ProductNotExistsException,
    ErrorGettingProductData,
)

_product_price = {}
product_by_barcode = {}
products_cache = None
last_cache_time = None
lock = Lock()
last_snapshot_id = None
logger = logging.getLogger("PosActions")

mbcontext = pyscripts.mbcontext  # type: MBEasyContext


@action
def getProducts(pos_id, force=False, snapshot_id=None):
    global products_cache
    global last_cache_time
    global last_snapshot_id
    global lock
    global product_by_barcode
    force = str(force).lower() == "true"
    seconds_to_renew_cache = 3600
    if lock.locked():
        raise StopAction()
    with lock:
        now = datetime.now()
        seconds_to_renew = _get_seconds_to_renew(seconds_to_renew_cache) if not force else 0
        reached_cache_time = _reached_seconds_to_renew_cache(now, seconds_to_renew)
        changed_last_snapshot_id = snapshot_id is not None and last_snapshot_id != snapshot_id
        if not products_cache or reached_cache_time or changed_last_snapshot_id:
            last_cache_time = now
            products = json.loads(_get_products(pos_id=pos_id))
            products_cache = json.dumps(products[0])
            _load_product_prices(products[0])
            _load_product_by_barcode(products[1])

            if snapshot_id is not None:
                last_snapshot_id = snapshot_id

    return products_cache


def _load_product_prices(
    products_dict,       # type: Dict
):
    # type: (...) -> None

    for key, value in products_dict.items():
        prices = value["price"]
        if not prices:
            continue

        _product_price[key] = prices.values()[0]


def _load_product_by_barcode(
    barcode_dict,       # type: Dict
):
    # type: (...) -> None

    for key, value in barcode_dict.items():
        product_by_barcode[int(key)] = value


def _get_seconds_to_renew(seconds_to_renew_cache):
    seconds_to_renew = 180
    try:
        seconds_to_renew = int(seconds_to_renew_cache)
    except (TypeError, ValueError) as _:
        pass
    return seconds_to_renew


def _reached_seconds_to_renew_cache(now, seconds_to_renew_cache):
    return not last_cache_time or now > last_cache_time + timedelta(seconds=seconds_to_renew_cache)


def _get_products(pos_id):
    msg = mbcontext.MB_EasySendMessage(
        dest_name='ORDERMGR{}'.format(pos_id),
        token=TK_SLCTRL_PMGR_GET_PRODUCTS,
        format=FM_PARAM,
        data=pos_id,
    )
    if msg.token != TK_SYS_ACK:
        raise ErrorGettingProductData(msg.data)
    if msg.token == TK_SYS_ACK:
        return msg.data


def get_product_by_barcode(barcode):
    if barcode and int(barcode) in product_by_barcode:
        return product_by_barcode[int(barcode)]

    return None


def get_product_price_from_cache(plu, pos_id):
    global _product_price

    if _product_price == {}:
        getProducts(pos_id)

    if str(plu) not in _product_price:
        raise ProductNotExistsException(str(plu))

    return _product_price[str(plu)]



from logging import Logger

from mw_helper import convert_to_dict, ensure_list
from production.box import ProductionBox
from production.model import ProductionOrder
from typing import List, Union

from ._OrderChangerProductionBox import OrderChangerProductionBox


class DoneWhenAnyDoneBox(OrderChangerProductionBox):
    def __init__(self, name, sons, part_codes, logger=None):
        # type: (str, Union[List[ProductionBox], ProductionBox], Union[str, List[str]], Logger) -> None
        super(DoneWhenAnyDoneBox, self).__init__(name, sons, logger)
        self.part_codes = convert_to_dict(ensure_list(part_codes))

    def change_order(self, order):
        if self.has_done_items(order):
            tracked_items = self.get_tracked_items(order)
            self.add_done_on_tracked_items(tracked_items)

        return order

    def has_done_items(self, order):
        # type: (ProductionOrder) -> bool
        for item in order.items:
            if self.has_any_done(item):
                return True

        return False

    def has_any_done(self, item):
        if item.part_code not in self.part_codes and item.has_tag("done") or item.has_tag("served"):
            return True
        for son in item.items:
            if self.has_any_done(son):
                return True

        return False

    def get_tracked_items(self, order, tracked_items=None):
        if tracked_items is None:
            tracked_items = []
        for item in order.items:
            if item.part_code in self.part_codes:
                tracked_items.append(item)
            else:
                for son in item.items:
                    self.get_tracked_items(son, tracked_items)

        return tracked_items

    @staticmethod
    def add_done_on_tracked_items(tracked_items):
        for item in tracked_items:
            if item.does_not_have_tag("done") and item.does_not_have_tag("served"):
                item.add_tag("done")



# vatin.py - function to validate any given VATIN.
#
# Copyright (C) 2020 Leandro Regueiro
# Copyright (C) 2021 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""VATIN (International value added tax identification number)

The number VAT identification number (VATIN) is an identifier used in many
countries. It starts with an ISO 3166-1 alpha-2 (2 letters) country code
(except for Greece, which uses EL, instead of GR) and is followed by the
country-specific the identifier.

This module supports all VAT numbers that are supported in python-stdnum.

More information:

* https://en.wikipedia.org/wiki/VAT_identification_number

>>> validate('FR 40 303 265 045')
'FR40303265045'
>>> validate('DE136,695 976')
'DE136695976'
>>> validate('BR16.727.230/0001-97')
'BR16727230000197'
>>> validate('FR 40 303')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> validate('XX')
Traceback (most recent call last):
    ...
InvalidComponent: ...
"""

import re

from stdnum.exceptions import *
from stdnum.util import clean, get_cc_module


# Cache of country code modules
_country_modules = dict()


def _get_cc_module(cc):
    """Get the VAT number module based on the country code."""
    # Greece uses a "wrong" country code, special case for Northern Ireland
    cc = cc.lower().replace('el', 'gr').replace('xi', 'gb')
    if not re.match(r'^[a-z]{2}$', cc):
        raise InvalidFormat()
    if cc not in _country_modules:
        _country_modules[cc] = get_cc_module(cc, 'vat')
    if not _country_modules[cc]:
        raise InvalidComponent()  # unknown/unsupported country code
    return _country_modules[cc]


def compact(number):
    """Convert the number to the minimal representation."""
    number = clean(number).strip()
    module = _get_cc_module(number[:2])
    return number[:2] + module.compact(number[2:])


def validate(number):
    """Check if the number is a valid VAT number.

    This performs the country-specific check for the number.
    """
    number = clean(number, '').strip()
    module = _get_cc_module(number[:2])
    try:
        return number[:2].upper() + module.validate(number[2:])
    except ValidationError:
        return module.validate(number)


def is_valid(number):
    """Check if the number is a valid VAT number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# coding:utf-8
import datetime
import functools
import time
from datetime import timedelta

from backoff._common import (_init_wait_gen, _maybe_call, _next_wait)


def _call_handlers(hdlrs, target, args, kwargs, tries, elapsed, **extra):
    details = {
        'target': target,
        'args': args,
        'kwargs': kwargs,
        'tries': tries,
        'elapsed': elapsed,
    }
    details.update(extra)
    for hdlr in hdlrs:
        hdlr(details)


def retry_predicate(target, wait_gen, predicate,
                    *,
                    max_tries, max_time, jitter,
                    on_success, on_backoff, on_giveup,
                    wait_gen_kwargs):

    @functools.wraps(target)
    def retry(*args, **kwargs):
        max_tries_value = _maybe_call(max_tries)
        max_time_value = _maybe_call(max_time)

        tries = 0
        start = datetime.datetime.now()
        wait = _init_wait_gen(wait_gen, wait_gen_kwargs)
        while True:
            tries += 1
            elapsed = timedelta.total_seconds(datetime.datetime.now() - start)
            details = {
                "target": target,
                "args": args,
                "kwargs": kwargs,
                "tries": tries,
                "elapsed": elapsed,
            }

            ret = target(*args, **kwargs)
            if predicate(ret):
                max_tries_exceeded = (tries == max_tries_value)
                max_time_exceeded = (max_time_value is not None and
                                     elapsed >= max_time_value)

                if max_tries_exceeded or max_time_exceeded:
                    _call_handlers(on_giveup, **details, value=ret)
                    break

                try:
                    seconds = _next_wait(wait, ret, jitter, elapsed,
                                         max_time_value)
                except StopIteration:
                    _call_handlers(on_giveup, **details)
                    break

                _call_handlers(on_backoff, **details,
                               value=ret, wait=seconds)

                time.sleep(seconds)
                continue
            else:
                _call_handlers(on_success, **details, value=ret)
                break

        return ret

    return retry


def retry_exception(target, wait_gen, exception,
                    *,
                    max_tries, max_time, jitter, giveup,
                    on_success, on_backoff, on_giveup, raise_on_giveup,
                    wait_gen_kwargs):

    @functools.wraps(target)
    def retry(*args, **kwargs):
        max_tries_value = _maybe_call(max_tries)
        max_time_value = _maybe_call(max_time)

        tries = 0
        start = datetime.datetime.now()
        wait = _init_wait_gen(wait_gen, wait_gen_kwargs)
        while True:
            tries += 1
            elapsed = timedelta.total_seconds(datetime.datetime.now() - start)
            details = {
                "target": target,
                "args": args,
                "kwargs": kwargs,
                "tries": tries,
                "elapsed": elapsed,
            }

            try:
                ret = target(*args, **kwargs)
            except exception as e:
                max_tries_exceeded = (tries == max_tries_value)
                max_time_exceeded = (max_time_value is not None and
                                     elapsed >= max_time_value)

                if giveup(e) or max_tries_exceeded or max_time_exceeded:
                    _call_handlers(on_giveup, **details, exception=e)
                    if raise_on_giveup:
                        raise
                    return None

                try:
                    seconds = _next_wait(wait, e, jitter, elapsed,
                                         max_time_value)
                except StopIteration:
                    _call_handlers(on_giveup, **details, exception=e)
                    raise e

                _call_handlers(on_backoff, **details, wait=seconds,
                               exception=e)

                time.sleep(seconds)
            else:
                _call_handlers(on_success, **details)

                return ret
    return retry



from datetime import datetime
import json
from typing import (
    Any,
    Dict,
    List,
    Optional,
)
from xml.etree import cElementTree as eTree

from sysactions import (
    action,
    get_user_information,
    show_any_dialog,
    show_confirmation,
    show_keyboard,
    show_messagebox,
    StopAction,
)

from mw_helper import show_editable_list_box_dialog, show_numpad_dialog

from actions.models import UserLevels
from actions.products import getProducts
from application.interactor import DailyGoalsInteractor
from model.customexception import (
    ComponentNotFound,
    DateFormatException,
    ErrorGettingSalesChallenge,
)
from manager.reports import sale_challenge_report


DATE_FORMAT = "%Y-%m-%d"


@action
def show_sales_challenges(
    pos_id,         # type: int
    operator_id,    # type: str
):
    # type: (...) -> None

    operator_level = eTree.XML(get_user_information(user_number=operator_id)).find("user").get("Level")
    is_manager = int(operator_level) >= UserLevels.MANAGER.value
    options_header = "$DESCRIPTION#$BEGIN_DATE#$END_DATE#$STATUS#$ACTIONS"
    options, sale_challenges = _get_sales_challenges(pos_id)
    options = options_header + "|" + "|".join(options)
    body = dict(isManager=is_manager, rows=options)
    response = show_editable_list_box_dialog(
        pos_id=pos_id,
        options=json.dumps(body).encode("utf-8"),
        title="$CHALLENGES",
        buttons="$CREATE_NEW|$CANCEL",
    )
    if not response:
        return
    action_name = response.split(',')[-1]
    if action_name == 'CREATE_NEW':
        _create_sale_challenge(
            pos_id=pos_id,
            operator_id=operator_id,
        )
    else:
        action_name, option_line = action_name.split(':')

        sale_challenge_id = sale_challenges[int(option_line) - 1]["challenge_id"]
        if action_name == 'EDIT':
            _edit_sale_challenge(
                pos_id=pos_id,
                sale_challenge=sale_challenges[int(option_line) - 1],
                operator_id=operator_id,
                editable=is_manager,
            )
        elif action_name == 'DELETE':
            _delete_sale_challenge(
                pos_id=pos_id,
                sale_challenge_id=sale_challenge_id,
                operator_id=operator_id,
            )
        elif action_name == 'REPORT':
            sale_challenge_report(
                pos_id=pos_id,
                sale_challenge_id=sale_challenge_id,
                is_manager=is_manager,
                operator_id=operator_id,
            )


def _display_and_validate_sale_challenge(pos_id, sale_challenge, can_edit=True, is_update=False):
    # type: (int, Dict, bool, bool) -> Any[Dict, None]

    sale_challenge["canEdit"] = can_edit
    response = show_any_dialog(pos_id, "editSaleChallengeDialog", "", "", json.dumps(sale_challenge), "", "", 600000,
                               "", "", "", "", "", False)

    sale_challenge_resp = json.loads(response)

    if not can_edit or not sale_challenge_resp:
        return None

    if is_update:
        is_ok, message = _validate_sale_challenge(sale_challenge_resp, sale_challenge)
    else:
        is_ok, message = _validate_sale_challenge(sale_challenge_resp)

    if not is_ok:
        show_messagebox(pos_id, message, "$ERROR")
        new_sale_challenge = dict(name=sale_challenge_resp["NAME"], startDate=sale_challenge_resp["START_DATE"],
                                  endDate=sale_challenge_resp["END_DATE"], products=sale_challenge_resp["PRODUCTS"])
        return _display_and_validate_sale_challenge(pos_id, new_sale_challenge, can_edit, is_update)
    return sale_challenge_resp


def _create_sale_challenge(
    pos_id,         # type: int
    operator_id,    # type: str
):
    # type: (...) -> None

    sale_challenge = dict(name="", startDate="", endDate="", products=[])

    sale_challenge_resp = _display_and_validate_sale_challenge(pos_id, sale_challenge)
    if sale_challenge_resp is None:
        return

    if _insert_sale_challenge(
        pos_id=pos_id,
        sale_challenge=sale_challenge_resp,
        operator_id=operator_id,
    ):
        show_messagebox(pos_id, "$SALE_CHALLENGE_CREATION_SUCCESS", "$INFORMATION")


def _edit_sale_challenge(
    pos_id,             # type: int
    sale_challenge,     # type: Dict
    operator_id,        # type: str
    editable=False,     # type: bool
):
    # type: (...) -> None

    all_products = json.loads(getProducts(pos_id=pos_id))
    products = json.loads(sale_challenge["products"])
    products = [str(p) for p in products if str(p) in all_products]
    sale_challenge_body = dict(name=sale_challenge["name"], startDate=sale_challenge["start_date"],
                               endDate=sale_challenge["end_date"], products=products)

    today = datetime.today().date()
    end_date = datetime.strptime(sale_challenge["end_date"], DATE_FORMAT).date()
    if today > end_date or not editable:
        sale_challenge_resp = _display_and_validate_sale_challenge(pos_id, sale_challenge_body, can_edit=False)
    else:
        sale_challenge_resp = _display_and_validate_sale_challenge(pos_id, sale_challenge_body, is_update=True)
    if sale_challenge_resp is None:
        return

    if _update_sale_challenge(
        sale_challenge=sale_challenge_resp,
        sale_challenge_id=sale_challenge["challenge_id"],
        operator_id=operator_id,
    ):
        show_messagebox(pos_id, "$SALE_CHALLENGE_EDITION_SUCCESS", "$INFORMATION")


def _delete_sale_challenge(
    pos_id,               # type: int
    sale_challenge_id,    # type: int
    operator_id,          # type: str
):
    # type: (...) -> None

    confirm = show_confirmation(pos_id, message="$DELETE_SALE_CHALLENGE_CONFIRMATION", buttons="$YES|$NO")
    if confirm:
        DailyGoalsInteractor.remove_sales_challenge(sale_challenge_id)
    else:
        show_sales_challenges(
            pos_id=pos_id,
            operator_id=operator_id,
        )


def _get_sales_challenges(pos_id):
    # type: (int) -> (List[str], List[Dict])

    try:
        options = []
        sales_challenges = DailyGoalsInteractor.get_sales_challenge(pos_id, 5)

        for challenge in sales_challenges:
            line = u"{}#{}#{}#{}".format(
                challenge["name"],
                challenge["start_date"],
                challenge["end_date"],
                challenge["challenge_id"],
            )
            options.append(line)

        return options, sales_challenges

    except (ComponentNotFound, ErrorGettingSalesChallenge) as ex:
        show_messagebox(
            posid=pos_id,
            message=ex.message,
            title="$ERROR",
            icon="error",
        )
        raise StopAction()


def _build_body_sale_challenge(
    sale_challenge,             # type: Dict
    operator_id,                # type: str
    sale_challenge_id=None,     # type: Optional[int]
):
    # type: (...) -> Dict

    user_id = int(operator_id)
    products = [int(p) for p in sale_challenge["PRODUCTS"]]
    return dict(challenge_id=sale_challenge_id, name=sale_challenge["NAME"],
                start_date=sale_challenge["START_DATE"], end_date=sale_challenge["END_DATE"],
                products=json.dumps(products), created_by=user_id)


def _insert_sale_challenge(
    pos_id,             # type: int
    sale_challenge,     # type: Dict
    operator_id,        # type: Optional[str]
):
    # type: (...) -> bool

    formatted_sale_challenge = _build_body_sale_challenge(
        sale_challenge=sale_challenge,
        operator_id=operator_id,
    )
    has_active_sale_challenges = DailyGoalsInteractor.get_sales_challenge(
        pos_id=pos_id,
        quantity=1,
    )
    if not has_active_sale_challenges:
        return DailyGoalsInteractor.insert_sales_challenge(sale_challenge=formatted_sale_challenge)
    else:
        show_messagebox(
            posid=pos_id,
            message="$ALREADY_ACTIVE_SALE_CHALLENGE",
        )


def _update_sale_challenge(
    sale_challenge,     # type: Dict
    sale_challenge_id,  # type: int
    operator_id,        # type: str
):
    # type: (...) -> bool

    formatted_sale_challenge = _build_body_sale_challenge(
        sale_challenge=sale_challenge,
        operator_id=operator_id,
        sale_challenge_id=sale_challenge_id,
    )
    return DailyGoalsInteractor.update_sales_challenge(formatted_sale_challenge)


@action
def show_challenge_keyboard_dialog(pos_id, message='', last_value=''):
    response = show_keyboard(pos_id, message, defvalue=last_value, title="", numpad=False)
    return response if response else last_value


@action
def show_challenge_numpad_dialog(pos_id, title='', last_value=''):
    return show_numpad_dialog(
        pos_id=pos_id,
        title=title,
        message="",
        mask="DATE",
        default_value=last_value,
    )


def _validate_sale_challenge(sale_challenge, original_sale_challenge=None):
    # type: (Dict, Dict) -> (bool, str)

    if "NAME" in sale_challenge and len(sale_challenge["NAME"]) == 0:
        return False, "$SALE_CHALLENGE_WITHOUT_NAME"
    if "PRODUCTS" in sale_challenge and len(sale_challenge["PRODUCTS"]) == 0:
        return False, "$SALE_CHALLENGE_WITHOUT_PRODUCTS"
    if "START_DATE" in sale_challenge and len(sale_challenge["START_DATE"]) == 0:
        return False, "$SALE_CHALLENGE_WITHOUT_START_DATE"
    if "END_DATE" in sale_challenge and len(sale_challenge["END_DATE"]) == 0:
        return False, "$SALE_CHALLENGE_WITHOUT_END_DATE"
    try:
        today = datetime.today().date()
        start_date = datetime.strptime(sale_challenge["START_DATE"], DATE_FORMAT).date()
        end_date = datetime.strptime(sale_challenge["END_DATE"], DATE_FORMAT).date()
        if today > start_date:
            return False, "$SALE_CHALLENGE_START_DATE_CANNOT_BE_EARLIER_THAN_TODAY"
        if today > end_date:
            return False, "$SALE_CHALLENGE_END_DATE_CANNOT_BE_EARLIER_THAN_TODAY"
        if start_date > end_date:
            return False, "$SALE_CHALLENGE_START_DATE_AFTER_THAN_END_DATE"
    except (DateFormatException,):
        return False, "$INVALID_DATE_FORMAT"

    if original_sale_challenge:
        if sale_challenge["NAME"] != original_sale_challenge["name"] or \
                sale_challenge["START_DATE"] != original_sale_challenge["startDate"] or \
                sale_challenge["END_DATE"] != original_sale_challenge["endDate"] or \
                sale_challenge["PRODUCTS"] != original_sale_challenge["products"]:
            return True, ""
        return False, "$SALE_CHALLENGE_WITHOUT_MODIFICATIONS"

    return True, ""



from ._UpdaterService import UpdaterService
from ._StaticConfigurationService import StaticConfigurationService
from ._LoaderUpdaterService import LoaderUpdaterService
from ._CommonService import CommonService
from ._ApacheUpdaterService import ApacheUpdaterService
from ._ChromeUpdaterService import ChromeUpdaterService
from ._StartStopUpdaterService import StartStopUpdaterService
from ._MediaUpdaterService import MediaUpdaterService
from ._JanitorService import JanitorService



from messagebus import TokenCreator, TokenPriority


class BusTokens(object):
    TK_AUDITLOG_RESYNC = TokenCreator.create_token(
        priority=TokenPriority.high,
        message_group="14",
        number="1",
    )
    TK_SALE_ORDER_MODIFIED = TokenCreator.create_token(
        priority=TokenPriority.low,
        message_group="A",
        number="32",
    )
    TK_AUDITLOG_GET_OPERATORS_LOGOUT_SESSIONS = TokenCreator.create_token(
        priority=TokenPriority.low,
        message_group="A",
        number="33",
    )



# asmens.py - functions for handling Lithuanian personal numbers
# coding: utf-8
#
# Copyright (C) 2019 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Asmens kodas (Lithuanian, personal numbers).

The Asmens kodas consists of 11 digits. The first digits denotes the gender
and birth century, the second through seventh denotes the birth date,
followed by a three-digit serial and a check digit.

More information:

* https://lt.wikipedia.org/wiki/Asmens_kodas
* https://en.wikipedia.org/wiki/National_identification_number#Lithuania

>>> validate('33309240064')
'33309240064'
>>> validate('33309240164')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
"""

from stdnum.ee.ik import calc_check_digit, get_birth_date
from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' ').strip()


def validate(number, validate_birth_date=True):
    """Check if the number provided is valid. This checks the length,
    formatting, embedded date and check digit."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) != 11:
        raise InvalidLength()
    if validate_birth_date and number[0] != '9':
        get_birth_date(number)
    if number[-1] != calc_check_digit(number):
        raise InvalidChecksum()
    return number


def is_valid(number, validate_birth_date=True):
    """Check if the number provided is valid. This checks the length,
    formatting, embedded date and check digit."""
    try:
        return bool(validate(number, validate_birth_date))
    except ValidationError:
        return False



from helper import translate_pos_label
from sysactions import (
    action,
    get_model,
    get_poslist,
    show_text_preview,
    translate_message,
    get_podtype,
)

from actions.pos.close_day import close_day

@action
def all_pos_status(
    pos_id,  # type: str
):
    # type: (...) -> bool

    while True:
        data = _generate_all_pos_data()
        result = show_text_preview(
            posid=pos_id,
            texts=data,
            title='$SELECT_ITEM',
            buttons='$OK|$CLOSE_POS',
            defvalue='',
            onthefly=False,
            timeout=120000,
        )
        if result is not None and len(result) >= 1:
            button_clicked = result[0]
            if button_clicked == '1':
                if len(result) >= 2:
                    pos_id_to_close = result[1]
                    close_day(pos_id_to_close, pos_id)
                else:
                    return False
            else:
                return True


def _generate_all_pos_data():
    data = []
    for pos_id_to_list in sorted(get_poslist()):
        model = get_model(pos_id_to_list)

        pod_type = get_podtype(model)
        if pos_id_to_list > 0 and pod_type not in ("TT", "DL"):
            operator_label = translate_message(
                model=model,
                message_key="OPERATOR",
            )
            name_label = translate_message(
                model=model,
                message_key="NAME",
            )
            status = model.find("PosState").get("state")
            if status == "UNDEFINED":
                status = "CLOSED"

            status_message = translate_message(
                model=model,
                message_key=status,
            )

            business_period = model.find("PosState").get("period")
            if business_period != "0":
                business_period = business_period[6:8] + "/" + business_period[4:6] + "/" + business_period[0:4]
            else:
                business_period = ""

            operators = ""
            for operator in model.find("Operators"):
                operator_id = operator.get("id")
                operator_name = operator.get("name")
                operator_status = translate_message(
                    model=model,
                    message_key=operator.get("state"),
                )
                operators += OPERATOR_INFO.format(
                    operator_label,
                    operator_id,
                    name_label,
                    operator_name,
                    operator_status,
                )
            message = POS_STATUS_INFO.format(
                translate_message(
                    model=model,
                    message_key=status,
                ),
                translate_message(
                    model=model,
                    message_key="BUSINESS_PERIOD",
                ),
                business_period,
                operators,
            )

            pos = translate_pos_label(model=model)
            status_title = "{}{} ({}) - {} - {}".format(pos, pos_id_to_list, pod_type, status_message, business_period)
            data.append([pos_id_to_list, status_title, message, ])
    return data


OPERATOR_INFO = """
    {}:
        ID: {}
        {}: {}
        Status: {}

"""

POS_STATUS_INFO = """
Status: {}
{}: {}
{}
"""




class Statistics(object):
    def __init__(self, value):
        # type: (unicode) -> None
        values = value.split("|")
        self.name = values[0]
        self.format = values[1] if len(values) > 1 else ""



from msgbus import TK_SYS_ACK
from sysactions import (
    action,
    send_message,
    show_messagebox,
)
from mw_helper import show_numpad_dialog
from bustoken import (
    create_token,
    MSGPRT_LOW,
)


TK_PUMP_RETRANSMIT_ORDERS = create_token(
    priority=MSGPRT_LOW,
    message_grourp="B",
    number="1",
)


@action
def retransmit_orders(
    pos_id,    # type: str
):
    # type: (...) -> None

    order_id = show_numpad_dialog(
        pos_id=pos_id,
        title="$INSERT_ORDER_ID",
        message="",
        mask="",
    )
    if order_id is None:
        return

    reply = send_message(
        dest_name="bohpump",
        token=TK_PUMP_RETRANSMIT_ORDERS,
        data=str(order_id),
        timeout=30000000,
    )
    if reply.token == TK_SYS_ACK:
        show_messagebox(
            posid=pos_id,
            message="$RETRANSMIT_ORDER_SUCCESS",
        )
        return

    show_messagebox(
        posid=pos_id,
        message="$RETRANSMIT_ORDER_ERROR",
    )



import json
from typing import Optional

from custom_sysactions import user_authorized_to_execute_action
from mw_helper import show_messagebox
from sysactions import action

from do_split_items import (
    filter_sale_line_list,
    validate_fraction_qty,
    split_items,
)


@action
@user_authorized_to_execute_action
def do_split_items_equally(
    pos_id,            # type: str
    sale_line_json,    # type: json
):
    # type: (...) -> Optional[bool]

    from msgbusboundary import MsgBusTableService as tableService

    sale_line_list = filter_sale_line_list(
        pos_id=pos_id,
        sale_line_json=sale_line_json,
    )

    if not sale_line_list:
        return False

    table_id = tableService().get_current_table_id(pos_id=pos_id)
    table_seats = tableService().get_service_seats(
        pos_id=pos_id,
        table_id=table_id,
    )

    if validate_fraction_qty(
        pos_id=pos_id,
        fraction_qty=table_seats,
        message='$CANT_SPLIT_ITEMS_IN_ONE_SEAT_TABLE',
    ) is False:
        return False

    result = show_messagebox(
        posid=pos_id,
        message="$DISCOUNTS_AND_PAYMENTS_WONT_BE_DISTRIBUTED",
        buttons="$OK|$CANCEL",
    )
    if result == 1:
        return False

    split_items(
        pos_id=pos_id,
        table_id=table_id,
        sale_line_list=sale_line_list,
        fraction_qty=table_seats,
        distribute_items_in_seats=True,
    )



# orm/bulk_persistence.py
# Copyright (C) 2005-2023 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php
# mypy: ignore-errors


"""additional ORM persistence classes related to "bulk" operations,
specifically outside of the flush() process.

"""

from __future__ import annotations

from typing import Any
from typing import cast
from typing import Dict
from typing import Iterable
from typing import Optional
from typing import overload
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import Union

from . import attributes
from . import context
from . import evaluator
from . import exc as orm_exc
from . import loading
from . import persistence
from .base import NO_VALUE
from .context import AbstractORMCompileState
from .context import FromStatement
from .context import ORMFromStatementCompileState
from .context import QueryContext
from .. import exc as sa_exc
from .. import util
from ..engine import Dialect
from ..engine import result as _result
from ..sql import coercions
from ..sql import dml
from ..sql import expression
from ..sql import roles
from ..sql import select
from ..sql import sqltypes
from ..sql.base import _entity_namespace_key
from ..sql.base import CompileState
from ..sql.base import Options
from ..sql.dml import DeleteDMLState
from ..sql.dml import InsertDMLState
from ..sql.dml import UpdateDMLState
from ..util import EMPTY_DICT
from ..util.typing import Literal

if TYPE_CHECKING:
    from ._typing import DMLStrategyArgument
    from ._typing import OrmExecuteOptionsParameter
    from ._typing import SynchronizeSessionArgument
    from .mapper import Mapper
    from .session import _BindArguments
    from .session import ORMExecuteState
    from .session import Session
    from .session import SessionTransaction
    from .state import InstanceState
    from ..engine import Connection
    from ..engine import cursor
    from ..engine.interfaces import _CoreAnyExecuteParams

_O = TypeVar("_O", bound=object)


@overload
def _bulk_insert(
    mapper: Mapper[_O],
    mappings: Union[Iterable[InstanceState[_O]], Iterable[Dict[str, Any]]],
    session_transaction: SessionTransaction,
    isstates: bool,
    return_defaults: bool,
    render_nulls: bool,
    use_orm_insert_stmt: Literal[None] = ...,
    execution_options: Optional[OrmExecuteOptionsParameter] = ...,
) -> None:
    ...


@overload
def _bulk_insert(
    mapper: Mapper[_O],
    mappings: Union[Iterable[InstanceState[_O]], Iterable[Dict[str, Any]]],
    session_transaction: SessionTransaction,
    isstates: bool,
    return_defaults: bool,
    render_nulls: bool,
    use_orm_insert_stmt: Optional[dml.Insert] = ...,
    execution_options: Optional[OrmExecuteOptionsParameter] = ...,
) -> cursor.CursorResult[Any]:
    ...


def _bulk_insert(
    mapper: Mapper[_O],
    mappings: Union[Iterable[InstanceState[_O]], Iterable[Dict[str, Any]]],
    session_transaction: SessionTransaction,
    isstates: bool,
    return_defaults: bool,
    render_nulls: bool,
    use_orm_insert_stmt: Optional[dml.Insert] = None,
    execution_options: Optional[OrmExecuteOptionsParameter] = None,
) -> Optional[cursor.CursorResult[Any]]:
    base_mapper = mapper.base_mapper

    if session_transaction.session.connection_callable:
        raise NotImplementedError(
            "connection_callable / per-instance sharding "
            "not supported in bulk_insert()"
        )

    if isstates:
        if return_defaults:
            states = [(state, state.dict) for state in mappings]
            mappings = [dict_ for (state, dict_) in states]
        else:
            mappings = [state.dict for state in mappings]
    else:
        mappings = [dict(m) for m in mappings]
        _expand_composites(mapper, mappings)

    connection = session_transaction.connection(base_mapper)

    return_result: Optional[cursor.CursorResult[Any]] = None

    mappers_to_run = [
        (table, mp)
        for table, mp in base_mapper._sorted_tables.items()
        if table in mapper._pks_by_table
    ]

    if return_defaults:
        # not used by new-style bulk inserts, only used for legacy
        bookkeeping = True
    elif len(mappers_to_run) > 1:
        # if we have more than one table, mapper to run where we will be
        # either horizontally splicing, or copying values between tables,
        # we need the "bookkeeping" / deterministic returning order
        bookkeeping = True
    else:
        bookkeeping = False

    for table, super_mapper in mappers_to_run:
        # find bindparams in the statement. For bulk, we don't really know if
        # a key in the params applies to a different table since we are
        # potentially inserting for multiple tables here; looking at the
        # bindparam() is a lot more direct.   in most cases this will
        # use _generate_cache_key() which is memoized, although in practice
        # the ultimate statement that's executed is probably not the same
        # object so that memoization might not matter much.
        extra_bp_names = (
            [
                b.key
                for b in use_orm_insert_stmt._get_embedded_bindparams()
                if b.key in mappings[0]
            ]
            if use_orm_insert_stmt is not None
            else ()
        )

        records = (
            (
                None,
                state_dict,
                params,
                mapper,
                connection,
                value_params,
                has_all_pks,
                has_all_defaults,
            )
            for (
                state,
                state_dict,
                params,
                mp,
                conn,
                value_params,
                has_all_pks,
                has_all_defaults,
            ) in persistence._collect_insert_commands(
                table,
                ((None, mapping, mapper, connection) for mapping in mappings),
                bulk=True,
                return_defaults=bookkeeping,
                render_nulls=render_nulls,
                include_bulk_keys=extra_bp_names,
            )
        )

        result = persistence._emit_insert_statements(
            base_mapper,
            None,
            super_mapper,
            table,
            records,
            bookkeeping=bookkeeping,
            use_orm_insert_stmt=use_orm_insert_stmt,
            execution_options=execution_options,
        )
        if use_orm_insert_stmt is not None:
            if not use_orm_insert_stmt._returning or return_result is None:
                return_result = result
            elif result.returns_rows:
                assert bookkeeping
                return_result = return_result.splice_horizontally(result)

    if return_defaults and isstates:
        identity_cls = mapper._identity_class
        identity_props = [p.key for p in mapper._identity_key_props]
        for state, dict_ in states:
            state.key = (
                identity_cls,
                tuple([dict_[key] for key in identity_props]),
            )

    if use_orm_insert_stmt is not None:
        assert return_result is not None
        return return_result


@overload
def _bulk_update(
    mapper: Mapper[Any],
    mappings: Union[Iterable[InstanceState[_O]], Iterable[Dict[str, Any]]],
    session_transaction: SessionTransaction,
    isstates: bool,
    update_changed_only: bool,
    use_orm_update_stmt: Literal[None] = ...,
    enable_check_rowcount: bool = True,
) -> None:
    ...


@overload
def _bulk_update(
    mapper: Mapper[Any],
    mappings: Union[Iterable[InstanceState[_O]], Iterable[Dict[str, Any]]],
    session_transaction: SessionTransaction,
    isstates: bool,
    update_changed_only: bool,
    use_orm_update_stmt: Optional[dml.Update] = ...,
    enable_check_rowcount: bool = True,
) -> _result.Result[Any]:
    ...


def _bulk_update(
    mapper: Mapper[Any],
    mappings: Union[Iterable[InstanceState[_O]], Iterable[Dict[str, Any]]],
    session_transaction: SessionTransaction,
    isstates: bool,
    update_changed_only: bool,
    use_orm_update_stmt: Optional[dml.Update] = None,
    enable_check_rowcount: bool = True,
) -> Optional[_result.Result[Any]]:
    base_mapper = mapper.base_mapper

    search_keys = mapper._primary_key_propkeys
    if mapper._version_id_prop:
        search_keys = {mapper._version_id_prop.key}.union(search_keys)

    def _changed_dict(mapper, state):
        return {
            k: v
            for k, v in state.dict.items()
            if k in state.committed_state or k in search_keys
        }

    if isstates:
        if update_changed_only:
            mappings = [_changed_dict(mapper, state) for state in mappings]
        else:
            mappings = [state.dict for state in mappings]
    else:
        mappings = [dict(m) for m in mappings]
        _expand_composites(mapper, mappings)

    if session_transaction.session.connection_callable:
        raise NotImplementedError(
            "connection_callable / per-instance sharding "
            "not supported in bulk_update()"
        )

    connection = session_transaction.connection(base_mapper)

    # find bindparams in the statement. see _bulk_insert for similar
    # notes for the insert case
    extra_bp_names = (
        [
            b.key
            for b in use_orm_update_stmt._get_embedded_bindparams()
            if b.key in mappings[0]
        ]
        if use_orm_update_stmt is not None
        else ()
    )

    for table, super_mapper in base_mapper._sorted_tables.items():
        if not mapper.isa(super_mapper) or table not in mapper._pks_by_table:
            continue

        records = persistence._collect_update_commands(
            None,
            table,
            (
                (
                    None,
                    mapping,
                    mapper,
                    connection,
                    (
                        mapping[mapper._version_id_prop.key]
                        if mapper._version_id_prop
                        else None
                    ),
                )
                for mapping in mappings
            ),
            bulk=True,
            use_orm_update_stmt=use_orm_update_stmt,
            include_bulk_keys=extra_bp_names,
        )
        persistence._emit_update_statements(
            base_mapper,
            None,
            super_mapper,
            table,
            records,
            bookkeeping=False,
            use_orm_update_stmt=use_orm_update_stmt,
            enable_check_rowcount=enable_check_rowcount,
        )

    if use_orm_update_stmt is not None:
        return _result.null_result()


def _expand_composites(mapper, mappings):
    composite_attrs = mapper.composites
    if not composite_attrs:
        return

    composite_keys = set(composite_attrs.keys())
    populators = {
        key: composite_attrs[key]._populate_composite_bulk_save_mappings_fn()
        for key in composite_keys
    }
    for mapping in mappings:
        for key in composite_keys.intersection(mapping):
            populators[key](mapping)


class ORMDMLState(AbstractORMCompileState):
    is_dml_returning = True
    from_statement_ctx: Optional[ORMFromStatementCompileState] = None

    @classmethod
    def _get_orm_crud_kv_pairs(
        cls, mapper, statement, kv_iterator, needs_to_be_cacheable
    ):
        core_get_crud_kv_pairs = UpdateDMLState._get_crud_kv_pairs

        for k, v in kv_iterator:
            k = coercions.expect(roles.DMLColumnRole, k)

            if isinstance(k, str):
                desc = _entity_namespace_key(mapper, k, default=NO_VALUE)
                if desc is NO_VALUE:
                    yield (
                        coercions.expect(roles.DMLColumnRole, k),
                        coercions.expect(
                            roles.ExpressionElementRole,
                            v,
                            type_=sqltypes.NullType(),
                            is_crud=True,
                        )
                        if needs_to_be_cacheable
                        else v,
                    )
                else:
                    yield from core_get_crud_kv_pairs(
                        statement,
                        desc._bulk_update_tuples(v),
                        needs_to_be_cacheable,
                    )
            elif "entity_namespace" in k._annotations:
                k_anno = k._annotations
                attr = _entity_namespace_key(
                    k_anno["entity_namespace"], k_anno["proxy_key"]
                )
                yield from core_get_crud_kv_pairs(
                    statement,
                    attr._bulk_update_tuples(v),
                    needs_to_be_cacheable,
                )
            else:
                yield (
                    k,
                    v
                    if not needs_to_be_cacheable
                    else coercions.expect(
                        roles.ExpressionElementRole,
                        v,
                        type_=sqltypes.NullType(),
                        is_crud=True,
                    ),
                )

    @classmethod
    def _get_multi_crud_kv_pairs(cls, statement, kv_iterator):
        plugin_subject = statement._propagate_attrs["plugin_subject"]

        if not plugin_subject or not plugin_subject.mapper:
            return UpdateDMLState._get_multi_crud_kv_pairs(
                statement, kv_iterator
            )

        return [
            dict(
                cls._get_orm_crud_kv_pairs(
                    plugin_subject.mapper, statement, value_dict.items(), False
                )
            )
            for value_dict in kv_iterator
        ]

    @classmethod
    def _get_crud_kv_pairs(cls, statement, kv_iterator, needs_to_be_cacheable):
        assert (
            needs_to_be_cacheable
        ), "no test coverage for needs_to_be_cacheable=False"

        plugin_subject = statement._propagate_attrs["plugin_subject"]

        if not plugin_subject or not plugin_subject.mapper:
            return UpdateDMLState._get_crud_kv_pairs(
                statement, kv_iterator, needs_to_be_cacheable
            )

        return list(
            cls._get_orm_crud_kv_pairs(
                plugin_subject.mapper,
                statement,
                kv_iterator,
                needs_to_be_cacheable,
            )
        )

    @classmethod
    def get_entity_description(cls, statement):
        ext_info = statement.table._annotations["parententity"]
        mapper = ext_info.mapper
        if ext_info.is_aliased_class:
            _label_name = ext_info.name
        else:
            _label_name = mapper.class_.__name__

        return {
            "name": _label_name,
            "type": mapper.class_,
            "expr": ext_info.entity,
            "entity": ext_info.entity,
            "table": mapper.local_table,
        }

    @classmethod
    def get_returning_column_descriptions(cls, statement):
        def _ent_for_col(c):
            return c._annotations.get("parententity", None)

        def _attr_for_col(c, ent):
            if ent is None:
                return c
            proxy_key = c._annotations.get("proxy_key", None)
            if not proxy_key:
                return c
            else:
                return getattr(ent.entity, proxy_key, c)

        return [
            {
                "name": c.key,
                "type": c.type,
                "expr": _attr_for_col(c, ent),
                "aliased": ent.is_aliased_class,
                "entity": ent.entity,
            }
            for c, ent in [
                (c, _ent_for_col(c)) for c in statement._all_selected_columns
            ]
        ]

    def _setup_orm_returning(
        self,
        compiler,
        orm_level_statement,
        dml_level_statement,
        dml_mapper,
        *,
        use_supplemental_cols=True,
    ):
        """establish ORM column handlers for an INSERT, UPDATE, or DELETE
        which uses explicit returning().

        called within compilation level create_for_statement.

        The _return_orm_returning() method then receives the Result
        after the statement was executed, and applies ORM loading to the
        state that we first established here.

        """

        if orm_level_statement._returning:
            fs = FromStatement(
                orm_level_statement._returning,
                dml_level_statement,
                _adapt_on_names=False,
            )
            fs = fs.execution_options(**orm_level_statement._execution_options)
            fs = fs.options(*orm_level_statement._with_options)
            self.select_statement = fs
            self.from_statement_ctx = (
                fsc
            ) = ORMFromStatementCompileState.create_for_statement(fs, compiler)
            fsc.setup_dml_returning_compile_state(dml_mapper)

            dml_level_statement = dml_level_statement._generate()
            dml_level_statement._returning = ()

            cols_to_return = [c for c in fsc.primary_columns if c is not None]

            # since we are splicing result sets together, make sure there
            # are columns of some kind returned in each result set
            if not cols_to_return:
                cols_to_return.extend(dml_mapper.primary_key)

            if use_supplemental_cols:
                dml_level_statement = dml_level_statement.return_defaults(
                    # this is a little weird looking, but by passing
                    # primary key as the main list of cols, this tells
                    # return_defaults to omit server-default cols (and
                    # actually all cols, due to some weird thing we should
                    # clean up in crud.py).
                    # Since we have cols_to_return, just return what we asked
                    # for (plus primary key, which ORM persistence needs since
                    # we likely set bookkeeping=True here, which is another
                    # whole thing...).   We dont want to clutter the
                    # statement up with lots of other cols the user didn't
                    # ask for.  see #9685
                    *dml_mapper.primary_key,
                    supplemental_cols=cols_to_return,
                )
            else:
                dml_level_statement = dml_level_statement.returning(
                    *cols_to_return
                )

        return dml_level_statement

    @classmethod
    def _return_orm_returning(
        cls,
        session,
        statement,
        params,
        execution_options,
        bind_arguments,
        result,
    ):
        execution_context = result.context
        compile_state = execution_context.compiled.compile_state

        if (
            compile_state.from_statement_ctx
            and not compile_state.from_statement_ctx.compile_options._is_star
        ):
            load_options = execution_options.get(
                "_sa_orm_load_options", QueryContext.default_load_options
            )

            querycontext = QueryContext(
                compile_state.from_statement_ctx,
                compile_state.select_statement,
                params,
                session,
                load_options,
                execution_options,
                bind_arguments,
            )
            return loading.instances(result, querycontext)
        else:
            return result


class BulkUDCompileState(ORMDMLState):
    class default_update_options(Options):
        _dml_strategy: DMLStrategyArgument = "auto"
        _synchronize_session: SynchronizeSessionArgument = "auto"
        _can_use_returning: bool = False
        _is_delete_using: bool = False
        _is_update_from: bool = False
        _autoflush: bool = True
        _subject_mapper: Optional[Mapper[Any]] = None
        _resolved_values = EMPTY_DICT
        _eval_condition = None
        _matched_rows = None
        _identity_token = None

    @classmethod
    def can_use_returning(
        cls,
        dialect: Dialect,
        mapper: Mapper[Any],
        *,
        is_multitable: bool = False,
        is_update_from: bool = False,
        is_delete_using: bool = False,
        is_executemany: bool = False,
    ) -> bool:
        raise NotImplementedError()

    @classmethod
    def orm_pre_session_exec(
        cls,
        session,
        statement,
        params,
        execution_options,
        bind_arguments,
        is_pre_event,
    ):
        (
            update_options,
            execution_options,
        ) = BulkUDCompileState.default_update_options.from_execution_options(
            "_sa_orm_update_options",
            {
                "synchronize_session",
                "autoflush",
                "identity_token",
                "is_delete_using",
                "is_update_from",
                "dml_strategy",
            },
            execution_options,
            statement._execution_options,
        )
        bind_arguments["clause"] = statement
        try:
            plugin_subject = statement._propagate_attrs["plugin_subject"]
        except KeyError:
            assert False, "statement had 'orm' plugin but no plugin_subject"
        else:
            if plugin_subject:
                bind_arguments["mapper"] = plugin_subject.mapper
                update_options += {"_subject_mapper": plugin_subject.mapper}

        if "parententity" not in statement.table._annotations:
            update_options += {"_dml_strategy": "core_only"}
        elif not isinstance(params, list):
            if update_options._dml_strategy == "auto":
                update_options += {"_dml_strategy": "orm"}
            elif update_options._dml_strategy == "bulk":
                raise sa_exc.InvalidRequestError(
                    'Can\'t use "bulk" ORM insert strategy without '
                    "passing separate parameters"
                )
        else:
            if update_options._dml_strategy == "auto":
                update_options += {"_dml_strategy": "bulk"}

        sync = update_options._synchronize_session
        if sync is not None:
            if sync not in ("auto", "evaluate", "fetch", False):
                raise sa_exc.ArgumentError(
                    "Valid strategies for session synchronization "
                    "are 'auto', 'evaluate', 'fetch', False"
                )
            if update_options._dml_strategy == "bulk" and sync == "fetch":
                raise sa_exc.InvalidRequestError(
                    "The 'fetch' synchronization strategy is not available "
                    "for 'bulk' ORM updates (i.e. multiple parameter sets)"
                )

        if not is_pre_event:
            if update_options._autoflush:
                session._autoflush()

            if update_options._dml_strategy == "orm":
                if update_options._synchronize_session == "auto":
                    update_options = cls._do_pre_synchronize_auto(
                        session,
                        statement,
                        params,
                        execution_options,
                        bind_arguments,
                        update_options,
                    )
                elif update_options._synchronize_session == "evaluate":
                    update_options = cls._do_pre_synchronize_evaluate(
                        session,
                        statement,
                        params,
                        execution_options,
                        bind_arguments,
                        update_options,
                    )
                elif update_options._synchronize_session == "fetch":
                    update_options = cls._do_pre_synchronize_fetch(
                        session,
                        statement,
                        params,
                        execution_options,
                        bind_arguments,
                        update_options,
                    )
            elif update_options._dml_strategy == "bulk":
                if update_options._synchronize_session == "auto":
                    update_options += {"_synchronize_session": "evaluate"}

            # indicators from the "pre exec" step that are then
            # added to the DML statement, which will also be part of the cache
            # key.  The compile level create_for_statement() method will then
            # consume these at compiler time.
            statement = statement._annotate(
                {
                    "synchronize_session": update_options._synchronize_session,
                    "is_delete_using": update_options._is_delete_using,
                    "is_update_from": update_options._is_update_from,
                    "dml_strategy": update_options._dml_strategy,
                    "can_use_returning": update_options._can_use_returning,
                }
            )

        return (
            statement,
            util.immutabledict(execution_options).union(
                {"_sa_orm_update_options": update_options}
            ),
        )

    @classmethod
    def orm_setup_cursor_result(
        cls,
        session,
        statement,
        params,
        execution_options,
        bind_arguments,
        result,
    ):
        # this stage of the execution is called after the
        # do_orm_execute event hook.  meaning for an extension like
        # horizontal sharding, this step happens *within* the horizontal
        # sharding event handler which calls session.execute() re-entrantly
        # and will occur for each backend individually.
        # the sharding extension then returns its own merged result from the
        # individual ones we return here.

        update_options = execution_options["_sa_orm_update_options"]
        if update_options._dml_strategy == "orm":
            if update_options._synchronize_session == "evaluate":
                cls._do_post_synchronize_evaluate(
                    session, statement, result, update_options
                )
            elif update_options._synchronize_session == "fetch":
                cls._do_post_synchronize_fetch(
                    session, statement, result, update_options
                )
        elif update_options._dml_strategy == "bulk":
            if update_options._synchronize_session == "evaluate":
                cls._do_post_synchronize_bulk_evaluate(
                    session, params, result, update_options
                )
            return result

        return cls._return_orm_returning(
            session,
            statement,
            params,
            execution_options,
            bind_arguments,
            result,
        )

    @classmethod
    def _adjust_for_extra_criteria(cls, global_attributes, ext_info):
        """Apply extra criteria filtering.

        For all distinct single-table-inheritance mappers represented in the
        table being updated or deleted, produce additional WHERE criteria such
        that only the appropriate subtypes are selected from the total results.

        Additionally, add WHERE criteria originating from LoaderCriteriaOptions
        collected from the statement.

        """

        return_crit = ()

        adapter = ext_info._adapter if ext_info.is_aliased_class else None

        if (
            "additional_entity_criteria",
            ext_info.mapper,
        ) in global_attributes:
            return_crit += tuple(
                ae._resolve_where_criteria(ext_info)
                for ae in global_attributes[
                    ("additional_entity_criteria", ext_info.mapper)
                ]
                if ae.include_aliases or ae.entity is ext_info
            )

        if ext_info.mapper._single_table_criterion is not None:
            return_crit += (ext_info.mapper._single_table_criterion,)

        if adapter:
            return_crit = tuple(adapter.traverse(crit) for crit in return_crit)

        return return_crit

    @classmethod
    def _interpret_returning_rows(cls, mapper, rows):
        """translate from local inherited table columns to base mapper
        primary key columns.

        Joined inheritance mappers always establish the primary key in terms of
        the base table.   When we UPDATE a sub-table, we can only get
        RETURNING for the sub-table's columns.

        Here, we create a lookup from the local sub table's primary key
        columns to the base table PK columns so that we can get identity
        key values from RETURNING that's against the joined inheritance
        sub-table.

        the complexity here is to support more than one level deep of
        inheritance, where we have to link columns to each other across
        the inheritance hierarchy.

        """

        if mapper.local_table is not mapper.base_mapper.local_table:
            return rows

        # this starts as a mapping of
        # local_pk_col: local_pk_col.
        # we will then iteratively rewrite the "value" of the dict with
        # each successive superclass column
        local_pk_to_base_pk = {pk: pk for pk in mapper.local_table.primary_key}

        for mp in mapper.iterate_to_root():
            if mp.inherits is None:
                break
            elif mp.local_table is mp.inherits.local_table:
                continue

            t_to_e = dict(mp._table_to_equated[mp.inherits.local_table])
            col_to_col = {sub_pk: super_pk for super_pk, sub_pk in t_to_e[mp]}
            for pk, super_ in local_pk_to_base_pk.items():
                local_pk_to_base_pk[pk] = col_to_col[super_]

        lookup = {
            local_pk_to_base_pk[lpk]: idx
            for idx, lpk in enumerate(mapper.local_table.primary_key)
        }
        primary_key_convert = [
            lookup[bpk] for bpk in mapper.base_mapper.primary_key
        ]
        return [tuple(row[idx] for idx in primary_key_convert) for row in rows]

    @classmethod
    def _get_matched_objects_on_criteria(cls, update_options, states):
        mapper = update_options._subject_mapper
        eval_condition = update_options._eval_condition

        raw_data = [
            (state.obj(), state, state.dict)
            for state in states
            if state.mapper.isa(mapper) and not state.expired
        ]

        identity_token = update_options._identity_token
        if identity_token is not None:
            raw_data = [
                (obj, state, dict_)
                for obj, state, dict_ in raw_data
                if state.identity_token == identity_token
            ]

        result = []
        for obj, state, dict_ in raw_data:
            evaled_condition = eval_condition(obj)

            # caution: don't use "in ()" or == here, _EXPIRE_OBJECT
            # evaluates as True for all comparisons
            if (
                evaled_condition is True
                or evaled_condition is evaluator._EXPIRED_OBJECT
            ):
                result.append(
                    (
                        obj,
                        state,
                        dict_,
                        evaled_condition is evaluator._EXPIRED_OBJECT,
                    )
                )
        return result

    @classmethod
    def _eval_condition_from_statement(cls, update_options, statement):
        mapper = update_options._subject_mapper
        target_cls = mapper.class_

        evaluator_compiler = evaluator._EvaluatorCompiler(target_cls)
        crit = ()
        if statement._where_criteria:
            crit += statement._where_criteria

        global_attributes = {}
        for opt in statement._with_options:
            if opt._is_criteria_option:
                opt.get_global_criteria(global_attributes)

        if global_attributes:
            crit += cls._adjust_for_extra_criteria(global_attributes, mapper)

        if crit:
            eval_condition = evaluator_compiler.process(*crit)
        else:
            # workaround for mypy https://github.com/python/mypy/issues/14027
            def _eval_condition(obj):
                return True

            eval_condition = _eval_condition

        return eval_condition

    @classmethod
    def _do_pre_synchronize_auto(
        cls,
        session,
        statement,
        params,
        execution_options,
        bind_arguments,
        update_options,
    ):
        """setup auto sync strategy


        "auto" checks if we can use "evaluate" first, then falls back
        to "fetch"

        evaluate is vastly more efficient for the common case
        where session is empty, only has a few objects, and the UPDATE
        statement can potentially match thousands/millions of rows.

        OTOH more complex criteria that fails to work with "evaluate"
        we would hope usually correlates with fewer net rows.

        """

        try:
            eval_condition = cls._eval_condition_from_statement(
                update_options, statement
            )

        except evaluator.UnevaluatableError:
            pass
        else:
            return update_options + {
                "_eval_condition": eval_condition,
                "_synchronize_session": "evaluate",
            }

        update_options += {"_synchronize_session": "fetch"}
        return cls._do_pre_synchronize_fetch(
            session,
            statement,
            params,
            execution_options,
            bind_arguments,
            update_options,
        )

    @classmethod
    def _do_pre_synchronize_evaluate(
        cls,
        session,
        statement,
        params,
        execution_options,
        bind_arguments,
        update_options,
    ):
        try:
            eval_condition = cls._eval_condition_from_statement(
                update_options, statement
            )

        except evaluator.UnevaluatableError as err:
            raise sa_exc.InvalidRequestError(
                'Could not evaluate current criteria in Python: "%s". '
                "Specify 'fetch' or False for the "
                "synchronize_session execution option." % err
            ) from err

        return update_options + {
            "_eval_condition": eval_condition,
        }

    @classmethod
    def _get_resolved_values(cls, mapper, statement):
        if statement._multi_values:
            return []
        elif statement._ordered_values:
            return list(statement._ordered_values)
        elif statement._values:
            return list(statement._values.items())
        else:
            return []

    @classmethod
    def _resolved_keys_as_propnames(cls, mapper, resolved_values):
        values = []
        for k, v in resolved_values:
            if mapper and isinstance(k, expression.ColumnElement):
                try:
                    attr = mapper._columntoproperty[k]
                except orm_exc.UnmappedColumnError:
                    pass
                else:
                    values.append((attr.key, v))
            else:
                raise sa_exc.InvalidRequestError(
                    "Attribute name not found, can't be "
                    "synchronized back to objects: %r" % k
                )
        return values

    @classmethod
    def _do_pre_synchronize_fetch(
        cls,
        session,
        statement,
        params,
        execution_options,
        bind_arguments,
        update_options,
    ):
        mapper = update_options._subject_mapper

        select_stmt = (
            select(*(mapper.primary_key + (mapper.select_identity_token,)))
            .select_from(mapper)
            .options(*statement._with_options)
        )
        select_stmt._where_criteria = statement._where_criteria

        # conditionally run the SELECT statement for pre-fetch, testing the
        # "bind" for if we can use RETURNING or not using the do_orm_execute
        # event.  If RETURNING is available, the do_orm_execute event
        # will cancel the SELECT from being actually run.
        #
        # The way this is organized seems strange, why don't we just
        # call can_use_returning() before invoking the statement and get
        # answer?, why does this go through the whole execute phase using an
        # event?  Answer: because we are integrating with extensions such
        # as the horizontal sharding extention that "multiplexes" an individual
        # statement run through multiple engines, and it uses
        # do_orm_execute() to do that.

        can_use_returning = None

        def skip_for_returning(orm_context: ORMExecuteState) -> Any:
            bind = orm_context.session.get_bind(**orm_context.bind_arguments)
            nonlocal can_use_returning

            per_bind_result = cls.can_use_returning(
                bind.dialect,
                mapper,
                is_update_from=update_options._is_update_from,
                is_delete_using=update_options._is_delete_using,
                is_executemany=orm_context.is_executemany,
            )

            if can_use_returning is not None:
                if can_use_returning != per_bind_result:
                    raise sa_exc.InvalidRequestError(
                        "For synchronize_session='fetch', can't mix multiple "
                        "backends where some support RETURNING and others "
                        "don't"
                    )
            elif orm_context.is_executemany and not per_bind_result:
                raise sa_exc.InvalidRequestError(
                    "For synchronize_session='fetch', can't use multiple "
                    "parameter sets in ORM mode, which this backend does not "
                    "support with RETURNING"
                )
            else:
                can_use_returning = per_bind_result

            if per_bind_result:
                return _result.null_result()
            else:
                return None

        result = session.execute(
            select_stmt,
            params,
            execution_options=execution_options,
            bind_arguments=bind_arguments,
            _add_event=skip_for_returning,
        )
        matched_rows = result.fetchall()

        return update_options + {
            "_matched_rows": matched_rows,
            "_can_use_returning": can_use_returning,
        }


@CompileState.plugin_for("orm", "insert")
class BulkORMInsert(ORMDMLState, InsertDMLState):
    class default_insert_options(Options):
        _dml_strategy: DMLStrategyArgument = "auto"
        _render_nulls: bool = False
        _return_defaults: bool = False
        _subject_mapper: Optional[Mapper[Any]] = None
        _autoflush: bool = True
        _populate_existing: bool = False

    select_statement: Optional[FromStatement] = None

    @classmethod
    def orm_pre_session_exec(
        cls,
        session,
        statement,
        params,
        execution_options,
        bind_arguments,
        is_pre_event,
    ):
        (
            insert_options,
            execution_options,
        ) = BulkORMInsert.default_insert_options.from_execution_options(
            "_sa_orm_insert_options",
            {"dml_strategy", "autoflush", "populate_existing", "render_nulls"},
            execution_options,
            statement._execution_options,
        )
        bind_arguments["clause"] = statement
        try:
            plugin_subject = statement._propagate_attrs["plugin_subject"]
        except KeyError:
            assert False, "statement had 'orm' plugin but no plugin_subject"
        else:
            if plugin_subject:
                bind_arguments["mapper"] = plugin_subject.mapper
                insert_options += {"_subject_mapper": plugin_subject.mapper}

        if not params:
            if insert_options._dml_strategy == "auto":
                insert_options += {"_dml_strategy": "orm"}
            elif insert_options._dml_strategy == "bulk":
                raise sa_exc.InvalidRequestError(
                    'Can\'t use "bulk" ORM insert strategy without '
                    "passing separate parameters"
                )
        else:
            if insert_options._dml_strategy == "auto":
                insert_options += {"_dml_strategy": "bulk"}

        if insert_options._dml_strategy != "raw":
            # for ORM object loading, like ORMContext, we have to disable
            # result set adapt_to_context, because we will be generating a
            # new statement with specific columns that's cached inside of
            # an ORMFromStatementCompileState, which we will re-use for
            # each result.
            if not execution_options:
                execution_options = context._orm_load_exec_options
            else:
                execution_options = execution_options.union(
                    context._orm_load_exec_options
                )

        if not is_pre_event and insert_options._autoflush:
            session._autoflush()

        statement = statement._annotate(
            {"dml_strategy": insert_options._dml_strategy}
        )

        return (
            statement,
            util.immutabledict(execution_options).union(
                {"_sa_orm_insert_options": insert_options}
            ),
        )

    @classmethod
    def orm_execute_statement(
        cls,
        session: Session,
        statement: dml.Insert,
        params: _CoreAnyExecuteParams,
        execution_options: OrmExecuteOptionsParameter,
        bind_arguments: _BindArguments,
        conn: Connection,
    ) -> _result.Result:
        insert_options = execution_options.get(
            "_sa_orm_insert_options", cls.default_insert_options
        )

        if insert_options._dml_strategy not in (
            "raw",
            "bulk",
            "orm",
            "auto",
        ):
            raise sa_exc.ArgumentError(
                "Valid strategies for ORM insert strategy "
                "are 'raw', 'orm', 'bulk', 'auto"
            )

        result: _result.Result[Any]

        if insert_options._dml_strategy == "raw":
            result = conn.execute(
                statement, params or {}, execution_options=execution_options
            )
            return result

        if insert_options._dml_strategy == "bulk":
            mapper = insert_options._subject_mapper

            if (
                statement._post_values_clause is not None
                and mapper._multiple_persistence_tables
            ):
                raise sa_exc.InvalidRequestError(
                    "bulk INSERT with a 'post values' clause "
                    "(typically upsert) not supported for multi-table "
                    f"mapper {mapper}"
                )

            assert mapper is not None
            assert session._transaction is not None
            result = _bulk_insert(
                mapper,
                cast(
                    "Iterable[Dict[str, Any]]",
                    [params] if isinstance(params, dict) else params,
                ),
                session._transaction,
                isstates=False,
                return_defaults=insert_options._return_defaults,
                render_nulls=insert_options._render_nulls,
                use_orm_insert_stmt=statement,
                execution_options=execution_options,
            )
        elif insert_options._dml_strategy == "orm":
            result = conn.execute(
                statement, params or {}, execution_options=execution_options
            )
        else:
            raise AssertionError()

        if not bool(statement._returning):
            return result

        if insert_options._populate_existing:
            load_options = execution_options.get(
                "_sa_orm_load_options", QueryContext.default_load_options
            )
            load_options += {"_populate_existing": True}
            execution_options = execution_options.union(
                {"_sa_orm_load_options": load_options}
            )

        return cls._return_orm_returning(
            session,
            statement,
            params,
            execution_options,
            bind_arguments,
            result,
        )

    @classmethod
    def create_for_statement(cls, statement, compiler, **kw) -> BulkORMInsert:
        self = cast(
            BulkORMInsert,
            super().create_for_statement(statement, compiler, **kw),
        )

        if compiler is not None:
            toplevel = not compiler.stack
        else:
            toplevel = True
        if not toplevel:
            return self

        mapper = statement._propagate_attrs["plugin_subject"]
        dml_strategy = statement._annotations.get("dml_strategy", "raw")
        if dml_strategy == "bulk":
            self._setup_for_bulk_insert(compiler)
        elif dml_strategy == "orm":
            self._setup_for_orm_insert(compiler, mapper)

        return self

    @classmethod
    def _resolved_keys_as_col_keys(cls, mapper, resolved_value_dict):
        return {
            col.key if col is not None else k: v
            for col, k, v in (
                (mapper.c.get(k), k, v) for k, v in resolved_value_dict.items()
            )
        }

    def _setup_for_orm_insert(self, compiler, mapper):
        statement = orm_level_statement = cast(dml.Insert, self.statement)

        statement = self._setup_orm_returning(
            compiler,
            orm_level_statement,
            statement,
            dml_mapper=mapper,
            use_supplemental_cols=False,
        )
        self.statement = statement

    def _setup_for_bulk_insert(self, compiler):
        """establish an INSERT statement within the context of
        bulk insert.

        This method will be within the "conn.execute()" call that is invoked
        by persistence._emit_insert_statement().

        """
        statement = orm_level_statement = cast(dml.Insert, self.statement)
        an = statement._annotations

        emit_insert_table, emit_insert_mapper = (
            an["_emit_insert_table"],
            an["_emit_insert_mapper"],
        )

        statement = statement._clone()

        statement.table = emit_insert_table
        if self._dict_parameters:
            self._dict_parameters = {
                col: val
                for col, val in self._dict_parameters.items()
                if col.table is emit_insert_table
            }

        statement = self._setup_orm_returning(
            compiler,
            orm_level_statement,
            statement,
            dml_mapper=emit_insert_mapper,
            use_supplemental_cols=True,
        )

        if (
            self.from_statement_ctx is not None
            and self.from_statement_ctx.compile_options._is_star
        ):
            raise sa_exc.CompileError(
                "Can't use RETURNING * with bulk ORM INSERT.  "
                "Please use a different INSERT form, such as INSERT..VALUES "
                "or INSERT with a Core Connection"
            )

        self.statement = statement


@CompileState.plugin_for("orm", "update")
class BulkORMUpdate(BulkUDCompileState, UpdateDMLState):
    @classmethod
    def create_for_statement(cls, statement, compiler, **kw):
        self = cls.__new__(cls)

        dml_strategy = statement._annotations.get(
            "dml_strategy", "unspecified"
        )

        toplevel = not compiler.stack

        if toplevel and dml_strategy == "bulk":
            self._setup_for_bulk_update(statement, compiler)
        elif (
            dml_strategy == "core_only"
            or dml_strategy == "unspecified"
            and "parententity" not in statement.table._annotations
        ):
            UpdateDMLState.__init__(self, statement, compiler, **kw)
        elif not toplevel or dml_strategy in ("orm", "unspecified"):
            self._setup_for_orm_update(statement, compiler)

        return self

    def _setup_for_orm_update(self, statement, compiler, **kw):
        orm_level_statement = statement

        toplevel = not compiler.stack

        ext_info = statement.table._annotations["parententity"]

        self.mapper = mapper = ext_info.mapper

        self._resolved_values = self._get_resolved_values(mapper, statement)

        self._init_global_attributes(
            statement,
            compiler,
            toplevel=toplevel,
            process_criteria_for_toplevel=toplevel,
        )

        if statement._values:
            self._resolved_values = dict(self._resolved_values)

        new_stmt = statement._clone()

        # note if the statement has _multi_values, these
        # are passed through to the new statement, which will then raise
        # InvalidRequestError because UPDATE doesn't support multi_values
        # right now.
        if statement._ordered_values:
            new_stmt._ordered_values = self._resolved_values
        elif statement._values:
            new_stmt._values = self._resolved_values

        new_crit = self._adjust_for_extra_criteria(
            self.global_attributes, mapper
        )
        if new_crit:
            new_stmt = new_stmt.where(*new_crit)

        # if we are against a lambda statement we might not be the
        # topmost object that received per-execute annotations

        # do this first as we need to determine if there is
        # UPDATE..FROM

        UpdateDMLState.__init__(self, new_stmt, compiler, **kw)

        use_supplemental_cols = False

        if not toplevel:
            synchronize_session = None
        else:
            synchronize_session = compiler._annotations.get(
                "synchronize_session", None
            )
        can_use_returning = compiler._annotations.get(
            "can_use_returning", None
        )
        if can_use_returning is not False:
            # even though pre_exec has determined basic
            # can_use_returning for the dialect, if we are to use
            # RETURNING we need to run can_use_returning() at this level
            # unconditionally because is_delete_using was not known
            # at the pre_exec level
            can_use_returning = (
                synchronize_session == "fetch"
                and self.can_use_returning(
                    compiler.dialect, mapper, is_multitable=self.is_multitable
                )
            )

        if synchronize_session == "fetch" and can_use_returning:
            use_supplemental_cols = True

            # NOTE: we might want to RETURNING the actual columns to be
            # synchronized also.  however this is complicated and difficult
            # to align against the behavior of "evaluate".  Additionally,
            # in a large number (if not the majority) of cases, we have the
            # "evaluate" answer, usually a fixed value, in memory already and
            # there's no need to re-fetch the same value
            # over and over again.   so perhaps if it could be RETURNING just
            # the elements that were based on a SQL expression and not
            # a constant.   For now it doesn't quite seem worth it
            new_stmt = new_stmt.return_defaults(*new_stmt.table.primary_key)

        if toplevel:
            new_stmt = self._setup_orm_returning(
                compiler,
                orm_level_statement,
                new_stmt,
                dml_mapper=mapper,
                use_supplemental_cols=use_supplemental_cols,
            )

        self.statement = new_stmt

    def _setup_for_bulk_update(self, statement, compiler, **kw):
        """establish an UPDATE statement within the context of
        bulk insert.

        This method will be within the "conn.execute()" call that is invoked
        by persistence._emit_update_statement().

        """
        statement = cast(dml.Update, statement)
        an = statement._annotations

        emit_update_table, _ = (
            an["_emit_update_table"],
            an["_emit_update_mapper"],
        )

        statement = statement._clone()
        statement.table = emit_update_table

        UpdateDMLState.__init__(self, statement, compiler, **kw)

        if self._ordered_values:
            raise sa_exc.InvalidRequestError(
                "bulk ORM UPDATE does not support ordered_values() for "
                "custom UPDATE statements with bulk parameter sets.  Use a "
                "non-bulk UPDATE statement or use values()."
            )

        if self._dict_parameters:
            self._dict_parameters = {
                col: val
                for col, val in self._dict_parameters.items()
                if col.table is emit_update_table
            }
        self.statement = statement

    @classmethod
    def orm_execute_statement(
        cls,
        session: Session,
        statement: dml.Update,
        params: _CoreAnyExecuteParams,
        execution_options: OrmExecuteOptionsParameter,
        bind_arguments: _BindArguments,
        conn: Connection,
    ) -> _result.Result:
        update_options = execution_options.get(
            "_sa_orm_update_options", cls.default_update_options
        )

        if update_options._dml_strategy not in (
            "orm",
            "auto",
            "bulk",
            "core_only",
        ):
            raise sa_exc.ArgumentError(
                "Valid strategies for ORM UPDATE strategy "
                "are 'orm', 'auto', 'bulk', 'core_only'"
            )

        result: _result.Result[Any]

        if update_options._dml_strategy == "bulk":
            enable_check_rowcount = not statement._where_criteria

            assert update_options._synchronize_session != "fetch"

            if (
                statement._where_criteria
                and update_options._synchronize_session == "evaluate"
            ):
                raise sa_exc.InvalidRequestError(
                    "bulk synchronize of persistent objects not supported "
                    "when using bulk update with additional WHERE "
                    "criteria right now.  add synchronize_session=None "
                    "execution option to bypass synchronize of persistent "
                    "objects."
                )
            mapper = update_options._subject_mapper
            assert mapper is not None
            assert session._transaction is not None
            result = _bulk_update(
                mapper,
                cast(
                    "Iterable[Dict[str, Any]]",
                    [params] if isinstance(params, dict) else params,
                ),
                session._transaction,
                isstates=False,
                update_changed_only=False,
                use_orm_update_stmt=statement,
                enable_check_rowcount=enable_check_rowcount,
            )
            return cls.orm_setup_cursor_result(
                session,
                statement,
                params,
                execution_options,
                bind_arguments,
                result,
            )
        else:
            return super().orm_execute_statement(
                session,
                statement,
                params,
                execution_options,
                bind_arguments,
                conn,
            )

    @classmethod
    def can_use_returning(
        cls,
        dialect: Dialect,
        mapper: Mapper[Any],
        *,
        is_multitable: bool = False,
        is_update_from: bool = False,
        is_delete_using: bool = False,
        is_executemany: bool = False,
    ) -> bool:
        # normal answer for "should we use RETURNING" at all.
        normal_answer = (
            dialect.update_returning and mapper.local_table.implicit_returning
        )
        if not normal_answer:
            return False

        if is_executemany:
            return dialect.update_executemany_returning

        # these workarounds are currently hypothetical for UPDATE,
        # unlike DELETE where they impact MariaDB
        if is_update_from:
            return dialect.update_returning_multifrom

        elif is_multitable and not dialect.update_returning_multifrom:
            raise sa_exc.CompileError(
                f'Dialect "{dialect.name}" does not support RETURNING '
                "with UPDATE..FROM; for synchronize_session='fetch', "
                "please add the additional execution option "
                "'is_update_from=True' to the statement to indicate that "
                "a separate SELECT should be used for this backend."
            )

        return True

    @classmethod
    def _do_post_synchronize_bulk_evaluate(
        cls, session, params, result, update_options
    ):
        if not params:
            return

        mapper = update_options._subject_mapper
        pk_keys = [prop.key for prop in mapper._identity_key_props]

        identity_map = session.identity_map

        for param in params:
            identity_key = mapper.identity_key_from_primary_key(
                (param[key] for key in pk_keys),
                update_options._identity_token,
            )
            state = identity_map.fast_get_state(identity_key)
            if not state:
                continue

            evaluated_keys = set(param).difference(pk_keys)

            dict_ = state.dict
            # only evaluate unmodified attributes
            to_evaluate = state.unmodified.intersection(evaluated_keys)
            for key in to_evaluate:
                if key in dict_:
                    dict_[key] = param[key]

            state.manager.dispatch.refresh(state, None, to_evaluate)

            state._commit(dict_, list(to_evaluate))

            # attributes that were formerly modified instead get expired.
            # this only gets hit if the session had pending changes
            # and autoflush were set to False.
            to_expire = evaluated_keys.intersection(dict_).difference(
                to_evaluate
            )
            if to_expire:
                state._expire_attributes(dict_, to_expire)

    @classmethod
    def _do_post_synchronize_evaluate(
        cls, session, statement, result, update_options
    ):
        matched_objects = cls._get_matched_objects_on_criteria(
            update_options,
            session.identity_map.all_states(),
        )

        cls._apply_update_set_values_to_objects(
            session,
            update_options,
            statement,
            [(obj, state, dict_) for obj, state, dict_, _ in matched_objects],
        )

    @classmethod
    def _do_post_synchronize_fetch(
        cls, session, statement, result, update_options
    ):
        target_mapper = update_options._subject_mapper

        returned_defaults_rows = result.returned_defaults_rows
        if returned_defaults_rows:
            pk_rows = cls._interpret_returning_rows(
                target_mapper, returned_defaults_rows
            )

            matched_rows = [
                tuple(row) + (update_options._identity_token,)
                for row in pk_rows
            ]
        else:
            matched_rows = update_options._matched_rows

        objs = [
            session.identity_map[identity_key]
            for identity_key in [
                target_mapper.identity_key_from_primary_key(
                    list(primary_key),
                    identity_token=identity_token,
                )
                for primary_key, identity_token in [
                    (row[0:-1], row[-1]) for row in matched_rows
                ]
                if update_options._identity_token is None
                or identity_token == update_options._identity_token
            ]
            if identity_key in session.identity_map
        ]

        if not objs:
            return

        cls._apply_update_set_values_to_objects(
            session,
            update_options,
            statement,
            [
                (
                    obj,
                    attributes.instance_state(obj),
                    attributes.instance_dict(obj),
                )
                for obj in objs
            ],
        )

    @classmethod
    def _apply_update_set_values_to_objects(
        cls, session, update_options, statement, matched_objects
    ):
        """apply values to objects derived from an update statement, e.g.
        UPDATE..SET <values>

        """
        mapper = update_options._subject_mapper
        target_cls = mapper.class_
        evaluator_compiler = evaluator._EvaluatorCompiler(target_cls)
        resolved_values = cls._get_resolved_values(mapper, statement)
        resolved_keys_as_propnames = cls._resolved_keys_as_propnames(
            mapper, resolved_values
        )
        value_evaluators = {}
        for key, value in resolved_keys_as_propnames:
            try:
                _evaluator = evaluator_compiler.process(
                    coercions.expect(roles.ExpressionElementRole, value)
                )
            except evaluator.UnevaluatableError:
                pass
            else:
                value_evaluators[key] = _evaluator

        evaluated_keys = list(value_evaluators.keys())
        attrib = {k for k, v in resolved_keys_as_propnames}

        states = set()
        for obj, state, dict_ in matched_objects:
            to_evaluate = state.unmodified.intersection(evaluated_keys)

            for key in to_evaluate:
                if key in dict_:
                    # only run eval for attributes that are present.
                    dict_[key] = value_evaluators[key](obj)

            state.manager.dispatch.refresh(state, None, to_evaluate)

            state._commit(dict_, list(to_evaluate))

            # attributes that were formerly modified instead get expired.
            # this only gets hit if the session had pending changes
            # and autoflush were set to False.
            to_expire = attrib.intersection(dict_).difference(to_evaluate)
            if to_expire:
                state._expire_attributes(dict_, to_expire)

            states.add(state)
        session._register_altered(states)


@CompileState.plugin_for("orm", "delete")
class BulkORMDelete(BulkUDCompileState, DeleteDMLState):
    @classmethod
    def create_for_statement(cls, statement, compiler, **kw):
        self = cls.__new__(cls)

        dml_strategy = statement._annotations.get(
            "dml_strategy", "unspecified"
        )

        if (
            dml_strategy == "core_only"
            or dml_strategy == "unspecified"
            and "parententity" not in statement.table._annotations
        ):
            DeleteDMLState.__init__(self, statement, compiler, **kw)
            return self

        toplevel = not compiler.stack

        orm_level_statement = statement

        ext_info = statement.table._annotations["parententity"]
        self.mapper = mapper = ext_info.mapper

        self._init_global_attributes(
            statement,
            compiler,
            toplevel=toplevel,
            process_criteria_for_toplevel=toplevel,
        )

        new_stmt = statement._clone()

        new_crit = cls._adjust_for_extra_criteria(
            self.global_attributes, mapper
        )
        if new_crit:
            new_stmt = new_stmt.where(*new_crit)

        # do this first as we need to determine if there is
        # DELETE..FROM
        DeleteDMLState.__init__(self, new_stmt, compiler, **kw)

        use_supplemental_cols = False

        if not toplevel:
            synchronize_session = None
        else:
            synchronize_session = compiler._annotations.get(
                "synchronize_session", None
            )
        can_use_returning = compiler._annotations.get(
            "can_use_returning", None
        )
        if can_use_returning is not False:
            # even though pre_exec has determined basic
            # can_use_returning for the dialect, if we are to use
            # RETURNING we need to run can_use_returning() at this level
            # unconditionally because is_delete_using was not known
            # at the pre_exec level
            can_use_returning = (
                synchronize_session == "fetch"
                and self.can_use_returning(
                    compiler.dialect,
                    mapper,
                    is_multitable=self.is_multitable,
                    is_delete_using=compiler._annotations.get(
                        "is_delete_using", False
                    ),
                )
            )

        if can_use_returning:
            use_supplemental_cols = True

            new_stmt = new_stmt.return_defaults(*new_stmt.table.primary_key)

        if toplevel:
            new_stmt = self._setup_orm_returning(
                compiler,
                orm_level_statement,
                new_stmt,
                dml_mapper=mapper,
                use_supplemental_cols=use_supplemental_cols,
            )

        self.statement = new_stmt

        return self

    @classmethod
    def orm_execute_statement(
        cls,
        session: Session,
        statement: dml.Delete,
        params: _CoreAnyExecuteParams,
        execution_options: OrmExecuteOptionsParameter,
        bind_arguments: _BindArguments,
        conn: Connection,
    ) -> _result.Result:
        update_options = execution_options.get(
            "_sa_orm_update_options", cls.default_update_options
        )

        if update_options._dml_strategy == "bulk":
            raise sa_exc.InvalidRequestError(
                "Bulk ORM DELETE not supported right now. "
                "Statement may be invoked at the "
                "Core level using "
                "session.connection().execute(stmt, parameters)"
            )

        if update_options._dml_strategy not in ("orm", "auto", "core_only"):
            raise sa_exc.ArgumentError(
                "Valid strategies for ORM DELETE strategy are 'orm', 'auto', "
                "'core_only'"
            )

        return super().orm_execute_statement(
            session, statement, params, execution_options, bind_arguments, conn
        )

    @classmethod
    def can_use_returning(
        cls,
        dialect: Dialect,
        mapper: Mapper[Any],
        *,
        is_multitable: bool = False,
        is_update_from: bool = False,
        is_delete_using: bool = False,
        is_executemany: bool = False,
    ) -> bool:
        # normal answer for "should we use RETURNING" at all.
        normal_answer = (
            dialect.delete_returning and mapper.local_table.implicit_returning
        )
        if not normal_answer:
            return False

        # now get into special workarounds because MariaDB supports
        # DELETE...RETURNING but not DELETE...USING...RETURNING.
        if is_delete_using:
            # is_delete_using hint was passed.   use
            # additional dialect feature (True for PG, False for MariaDB)
            return dialect.delete_returning_multifrom

        elif is_multitable and not dialect.delete_returning_multifrom:
            # is_delete_using hint was not passed, but we determined
            # at compile time that this is in fact a DELETE..USING.
            # it's too late to continue since we did not pre-SELECT.
            # raise that we need that hint up front.

            raise sa_exc.CompileError(
                f'Dialect "{dialect.name}" does not support RETURNING '
                "with DELETE..USING; for synchronize_session='fetch', "
                "please add the additional execution option "
                "'is_delete_using=True' to the statement to indicate that "
                "a separate SELECT should be used for this backend."
            )

        return True

    @classmethod
    def _do_post_synchronize_evaluate(
        cls, session, statement, result, update_options
    ):
        matched_objects = cls._get_matched_objects_on_criteria(
            update_options,
            session.identity_map.all_states(),
        )

        to_delete = []

        for _, state, dict_, is_partially_expired in matched_objects:
            if is_partially_expired:
                state._expire(dict_, session.identity_map._modified)
            else:
                to_delete.append(state)

        if to_delete:
            session._remove_newly_deleted(to_delete)

    @classmethod
    def _do_post_synchronize_fetch(
        cls, session, statement, result, update_options
    ):
        target_mapper = update_options._subject_mapper

        returned_defaults_rows = result.returned_defaults_rows

        if returned_defaults_rows:
            pk_rows = cls._interpret_returning_rows(
                target_mapper, returned_defaults_rows
            )

            matched_rows = [
                tuple(row) + (update_options._identity_token,)
                for row in pk_rows
            ]
        else:
            matched_rows = update_options._matched_rows

        for row in matched_rows:
            primary_key = row[0:-1]
            identity_token = row[-1]

            # TODO: inline this and call remove_newly_deleted
            # once
            identity_key = target_mapper.identity_key_from_primary_key(
                list(primary_key),
                identity_token=identity_token,
            )
            if identity_key in session.identity_map:
                session._remove_newly_deleted(
                    [
                        attributes.instance_state(
                            session.identity_map[identity_key]
                        )
                    ]
                )



from xml.etree import cElementTree as Etree
from datetime import datetime

from model.exception import InvalidTableEventException
from model.log import LogEntry
from _EventProcessor import EventProcessor


def _get_log_entry_from_order_event(data, _):
    event = Etree.fromstring(data)
    table = _check_mandatory_information(event, data)

    created_at = datetime.strptime(table.get("startTS"), "%Y-%m-%d %H:%M:%S").strftime("%Y-%m-%dT%H:%M:%S.000")
    return LogEntry(
        pos_id=event.get("posId", -1),
        session_id=event.get("sessionId", ''),
        business_period=table.get("businessPeriod").replace('-', ''),
        event_type=event.get("subject"),
        created_at_gmt=created_at,
        event_data=Etree.tostring(table)
    )


def _check_mandatory_information(event, data):
    table = event.find(".//Table")
    if table is None:
        raise InvalidTableEventException(data)

    return table


class TableEventProcessor(EventProcessor):
    def parse_data(self, data, event_type):
        # type: (str, str) -> LogEntry

        return _get_log_entry_from_order_event(data, event_type)



import os

import cfgtools

hold_times = None


def get_hold_time_config():
    global hold_times

    if hold_times is not None:
        return hold_times

    config = cfgtools.read(os.environ["LOADERCFG"])
    hold_times = []
    hold_config = config.find_group("ProductionConfig")
    if hold_config is not None:
        config_hold_times = config.find_values("ProductionConfig.HoldTimes") or []
        for hold_time in config_hold_times:
            hold_times.append(str(int(int(hold_time) / 60)))

    return hold_times



# -*- coding: utf-8 -*-
import sysactions
import logging

from sysactions import (
    get_posfunction,
    get_podtype
)

from posot import OrderTakerException
from typing import (
    Any,
    Union,
)

from actions.order.must_sync_ui import must_sync_ui
from custom_sysactions import (
    user_authorized_to_execute_action,
    block_action_if_order_items_not_resolved,
    fix_current_order_in_model,
)

from actions.discount.clean_discounts import clean_order_payments
from actions.util import is_delivery_order
from actions.delivery.save_delivery_order import save_delivery_order

from model.customexception import OrderNotFoundException

logger = logging.getLogger("PosActions")


@sysactions.action
@user_authorized_to_execute_action
@block_action_if_order_items_not_resolved
def doStoreOrder(
    pos_id,                     	# type: Any[Union[str, int]]
    totalize="none",            	# type: str
    from_total_screen="false",  	# type: str
    pos_id_to_show_message=None,    # type: str
    show_popup=True,            	# type: bool
    *args                       	# type: Any
):
    # type: (...) -> None

    from posactions import (
        doTotal,
        fill_customer_properties_in_event,
    )
    if pos_id_to_show_message is None:
        pos_id_to_show_message = pos_id
    model = sysactions.get_model(pos_id)
    sysactions.check_current_order(pos_id, model=model, need_order=True)
    if from_total_screen.lower() == "true":
        must_sync_ui(
            pos_id=pos_id,
            different_than="TOTALED",
        )
    pos_ot = sysactions.get_posot(model)
    xml_order = sysactions.get_current_order(model)
    clean_order_payments(
        pos_id=pos_id,
        pos_id_to_show_message=pos_id_to_show_message,
        order=xml_order,
        check_manual_discounts=True,
        store=True,
    )

    if sysactions.get_current_order(model).get("type") != "SALE":
        sysactions.show_messagebox(
            posid=pos_id_to_show_message,
            message="$THIS_ORDER_TYPE_CANNOT_BE_STORED",
            icon="warning",
        )
        return

    is_dt_or_fc = get_podtype(model=model) in ["DT", "FC"]
    pod_function = get_posfunction(model=model) if is_dt_or_fc else get_podtype(model=model)
    if totalize == "none" and pod_function == "OT":
        totalize = "true"

    if totalize.lower() != "true" or pod_function == "OT":
        fill_customer_properties_in_event(
            pos_id=pos_id,
            pos_ot=pos_ot,
            pod_function=pod_function,
            key_event='onSave',
        )

    if totalize.lower() == "true" and doTotal(pos_id, screen_number="") == "Error":
        return

    order = model.find("CurrentOrder/Order")
    # Pre-venda
    pre_venda = None
    if pod_function == "OT":
        for prop in order.findall("CustomOrderProperties/OrderProperty"):
            key, value = prop.get("key"), prop.get("value")
            if key == "PRE_VENDA":
                pre_venda = value
                break
    try:
        pos_ot.storeOrder(pos_id)

        if pre_venda:
            sysactions.show_messagebox(
                posid=pos_id_to_show_message,
                message="Pre Venda: PV%010d" % int(pre_venda),
            )
        elif _save_order_on_remote_order(order, pos_id, pos_ot):
            if show_popup:
                sysactions.show_messagebox(
                    posid=pos_id_to_show_message,
                    message="$ORDER_SAVED",
                )

        else:
            sysactions.show_messagebox(
                posid=pos_id_to_show_message,
                message="$ERROR_SAVING_ORDER",
            )
    except OrderTakerException as ex:
        error_code = ex.getErrorCode()
        error_description = ex.getErrorDescr()
        if error_code == 1:
            order = model.find("CurrentOrder/Order")
            order_id = order.get("orderId")
            fix_current_order_in_model(
                pos_id=pos_id,
                model=model,
                order_taker=pos_ot,
                order_id=order_id,
            )
            raise OrderNotFoundException(
                error_code=error_code,
                message=error_description,
            )

        sysactions.show_messagebox(
            posid=pos_id_to_show_message,
            message="$ERROR_CODE_INFO|%d|%s" % (error_code, error_description),
        )

    order = model.find("CurrentOrder/Order")
    order_id = order.get("orderId")
    logger.debug("Pedido Salvo - Order %s - PosId %s" % (order_id, pos_id))


def _save_order_on_remote_order(order, pos_id, pos_ot):
    model = sysactions.get_model(pos_id)
    if is_delivery_order(model):
        order_id = order.get("orderId")

        try:
            save_delivery_order(order_id)
        except (Exception,):
            session = sysactions.get_operator_session(model)
            pos_ot.recallOrder(int(pos_id), int(order_id), session)
            return False

    return True



# testing/fixtures/orm.py
# Copyright (C) 2005-2023 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php
# mypy: ignore-errors
from __future__ import annotations

from typing import Any

import sqlalchemy as sa
from .base import TestBase
from .sql import TablesTest
from .. import assertions
from .. import config
from .. import schema
from ..entities import BasicEntity
from ..entities import ComparableEntity
from ..util import adict
from ... import orm
from ...orm import DeclarativeBase
from ...orm import events as orm_events
from ...orm import registry


class ORMTest(TestBase):
    @config.fixture
    def fixture_session(self):
        return fixture_session()


class MappedTest(ORMTest, TablesTest, assertions.AssertsExecutionResults):
    # 'once', 'each', None
    run_setup_classes = "once"

    # 'once', 'each', None
    run_setup_mappers = "each"

    classes: Any = None

    @config.fixture(autouse=True, scope="class")
    def _setup_tables_test_class(self):
        cls = self.__class__
        cls._init_class()

        if cls.classes is None:
            cls.classes = adict()

        cls._setup_once_tables()
        cls._setup_once_classes()
        cls._setup_once_mappers()
        cls._setup_once_inserts()

        yield

        cls._teardown_once_class()
        cls._teardown_once_metadata_bind()

    @config.fixture(autouse=True, scope="function")
    def _setup_tables_test_instance(self):
        self._setup_each_tables()
        self._setup_each_classes()
        self._setup_each_mappers()
        self._setup_each_inserts()

        yield

        orm.session.close_all_sessions()
        self._teardown_each_mappers()
        self._teardown_each_classes()
        self._teardown_each_tables()

    @classmethod
    def _teardown_once_class(cls):
        cls.classes.clear()

    @classmethod
    def _setup_once_classes(cls):
        if cls.run_setup_classes == "once":
            cls._with_register_classes(cls.setup_classes)

    @classmethod
    def _setup_once_mappers(cls):
        if cls.run_setup_mappers == "once":
            cls.mapper_registry, cls.mapper = cls._generate_registry()
            cls._with_register_classes(cls.setup_mappers)

    def _setup_each_mappers(self):
        if self.run_setup_mappers != "once":
            (
                self.__class__.mapper_registry,
                self.__class__.mapper,
            ) = self._generate_registry()

        if self.run_setup_mappers == "each":
            self._with_register_classes(self.setup_mappers)

    def _setup_each_classes(self):
        if self.run_setup_classes == "each":
            self._with_register_classes(self.setup_classes)

    @classmethod
    def _generate_registry(cls):
        decl = registry(metadata=cls._tables_metadata)
        return decl, decl.map_imperatively

    @classmethod
    def _with_register_classes(cls, fn):
        """Run a setup method, framing the operation with a Base class
        that will catch new subclasses to be established within
        the "classes" registry.

        """
        cls_registry = cls.classes

        class _Base:
            def __init_subclass__(cls) -> None:
                assert cls_registry is not None
                cls_registry[cls.__name__] = cls
                super().__init_subclass__()

        class Basic(BasicEntity, _Base):
            pass

        class Comparable(ComparableEntity, _Base):
            pass

        cls.Basic = Basic
        cls.Comparable = Comparable
        fn()

    def _teardown_each_mappers(self):
        # some tests create mappers in the test bodies
        # and will define setup_mappers as None -
        # clear mappers in any case
        if self.run_setup_mappers != "once":
            orm.clear_mappers()

    def _teardown_each_classes(self):
        if self.run_setup_classes != "once":
            self.classes.clear()

    @classmethod
    def setup_classes(cls):
        pass

    @classmethod
    def setup_mappers(cls):
        pass


class DeclarativeMappedTest(MappedTest):
    run_setup_classes = "once"
    run_setup_mappers = "once"

    @classmethod
    def _setup_once_tables(cls):
        pass

    @classmethod
    def _with_register_classes(cls, fn):
        cls_registry = cls.classes

        class _DeclBase(DeclarativeBase):
            __table_cls__ = schema.Table
            metadata = cls._tables_metadata
            type_annotation_map = {
                str: sa.String().with_variant(
                    sa.String(50), "mysql", "mariadb", "oracle"
                )
            }

            def __init_subclass__(cls, **kw) -> None:
                assert cls_registry is not None
                cls_registry[cls.__name__] = cls
                super().__init_subclass__(**kw)

        cls.DeclarativeBasic = _DeclBase

        # sets up cls.Basic which is helpful for things like composite
        # classes
        super()._with_register_classes(fn)

        if cls._tables_metadata.tables and cls.run_create_tables:
            cls._tables_metadata.create_all(config.db)


class RemoveORMEventsGlobally:
    @config.fixture(autouse=True)
    def _remove_listeners(self):
        yield
        orm_events.MapperEvents._clear()
        orm_events.InstanceEvents._clear()
        orm_events.SessionEvents._clear()
        orm_events.InstrumentationEvents._clear()
        orm_events.QueryEvents._clear()


_fixture_sessions = set()


def fixture_session(**kw):
    kw.setdefault("autoflush", True)
    kw.setdefault("expire_on_commit", True)

    bind = kw.pop("bind", config.db)

    sess = orm.Session(bind, **kw)
    _fixture_sessions.add(sess)
    return sess


def close_all_sessions():
    # will close all still-referenced sessions
    orm.close_all_sessions()
    _fixture_sessions.clear()


def stop_test_class_inside_fixtures(cls):
    close_all_sessions()
    orm.clear_mappers()


def after_test():
    if _fixture_sessions:
        close_all_sessions()



from _CalculateTaxesItemCommon import CalculateTaxesItemCommon
from _FiscalParameterController import FiscalParameterController, FiscalParameterNotFound
from _Utils import build_item_taxes
from _TipBuilderCommon import TipBuilderCommon
from _InfAdicBuilderCommon import InfAdicBuilderCommon
from _PrinterCommon import PrinterCommon



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=too-many-lines


class MetricInstruments:

    HTTP_SERVER_DURATION = "http.server.duration"

    HTTP_SERVER_REQUEST_SIZE = "http.server.request.size"

    HTTP_SERVER_RESPONSE_SIZE = "http.server.response.size"

    HTTP_SERVER_ACTIVE_REQUESTS = "http.server.active_requests"

    HTTP_CLIENT_DURATION = "http.client.duration"

    HTTP_CLIENT_REQUEST_SIZE = "http.client.request.size"

    HTTP_CLIENT_RESPONSE_SIZE = "http.client.response.size"

    DB_CLIENT_CONNECTIONS_USAGE = "db.client.connections.usage"



# coding=utf-8
import json
from logging import Logger

from helper import json_serialize
from requests import Response
from .._TranslatedMessage import TranslatedMessage


class ConfigurationReaderException(Exception):
    pass


class NotFound(Exception):
    pass


class ValidationError(Exception):
    pass


class ResponseError(Exception):
    def __init__(self, logger, response):
        # type: (Logger, Response) -> None

        self.logger = logger
        self.response = response
        self.error_status_code = response.status_code
        self.message = _get_response_content(response)
        response_str = json_serialize(response)

        message = "Unexpected response received. StatusCode: {}; Content: {}; FullResponse:\n{}"
        self.logger.error(message.format(self.error_status_code, self.message, response_str))

        super(ResponseError, self).__init__(self.message)


class NoResponseError(Exception):
    def __init__(self, logger, response):
        self.logger = logger
        self.response = response

        error_status_code = response.status_code
        content = _get_response_content(response)
        message = "No response. StatusCode: {}; Content: {}; FullResponse: {}"
        response_str = json_serialize(response)
        self.logger.error(message.format(error_status_code, content, response_str))
        super(NoResponseError, self).__init__(content)


class ForbiddenError(Exception):
    def __init__(self, logger, response):
        self.logger = logger
        self.response = response

        error_status_code = response.status_code
        content = _get_response_content(response)
        message = "Forbidden error. StatusCode: {}; Content: {}; FullResponse: {}"
        response_str = json_serialize(response)
        self.logger.error(message.format(error_status_code, content, response_str))
        super(ForbiddenError, self).__init__(content)


class NotFoundError(Exception):
    def __init__(self, logger, response):
        self.logger = logger
        self.response = response

        error_status_code = response.status_code
        content = _get_response_content(response)
        message = "Not found. StatusCode: {}; Content: {}; FullResponse: {}"
        response_str = json_serialize(response)
        self.logger.error(message.format(error_status_code, content, response_str))
        super(NotFoundError, self).__init__(content)


class NotAllowedError(Exception):
    def __init__(self, logger, response):
        self.logger = logger
        self.response = response

        error_status_code = response.status_code
        content = _get_response_content(response)
        message = "Method not allowed. StatusCode: {}; Content: {}; FullResponse: {}"
        response_str = json_serialize(response)
        self.logger.error(message.format(error_status_code, content, response_str))
        super(NotAllowedError, self).__init__(content)


class NotAvailableError(Exception):
    def __init__(self, logger, response):
        self.logger = logger
        self.response = response

        error_status_code = response.status_code
        content = _get_response_content(response)
        message = "Code not available. StatusCode: {}; Content: {}; FullResponse: {}"
        response_str = json_serialize(response)
        self.logger.error(message.format(error_status_code, content, response_str))
        super(NotAvailableError, self).__init__(content)


class InUseError(Exception):
    def __init__(self, logger, response):
        self.logger = logger
        self.response = response

        error_status_code = response.status_code
        content = _get_response_content(response)
        message = "Code in use on another terminal. StatusCode: {}; Content: {}; FullResponse: {}"
        response_str = json_serialize(response)
        self.logger.error(message.format(error_status_code, content, response_str))
        super(InUseError, self).__init__(content)


class AlreadyUnlockedError(Exception):
    def __init__(self, logger, response):
        self.logger = logger
        self.response = response

        error_status_code = response.status_code
        content = _get_response_content(response)
        message = "Code already unlocked. StatusCode: {}; Content: {}; FullResponse: {}"
        response_str = json_serialize(response)
        self.logger.error(message.format(error_status_code, content, response_str))
        super(AlreadyUnlockedError, self).__init__(content)


class ProcessorError(Exception):
    def __init__(self, logger, response):
        self.logger = logger
        self.response = response

        error_status_code = response.status_code
        content = _get_response_content(response)
        message = "Processing error. StatusCode: {}; Content: {}; FullResponse: {}"
        response_str = json_serialize(response)
        self.logger.error(message.format(error_status_code, content, response_str))
        super(ProcessorError, self).__init__(content)

    def __str__(self):
        return "ProcessorError. StatusCode: {}; Content: {}".format(
            self.response.status_code,
            self.response.content
        )

    def __repr__(self):
        return self.__str__()


class ClientCustomError(Exception):
    def __init__(
        self,
        message,   # type: TranslatedMessage
    ):
        # type: (...) -> None

        self.message = message


class GoneError(Exception):
    def __init__(self, *args, **kwargs):
        super(Exception, self).__init__()


class OrderHasNotBenefitListException(Exception):
    def __init__(
        self,       # type: OrderHasNotBenefitListException
        order_id,   # type: str
    ):
        # type: (...) -> None

        self.order_id = order_id

        super(Exception, self).__init__()

    def __str__(
        self,   # type: OrderHasNotBenefitListException
    ):
        # type: (...) -> str

        return "Order {} has not Benefit List".format(self.order_id)


class BenefitIdNotFoundInAppliedOrderBenefitsException(Exception):
    def __init__(
        self,           # type: BenefitIdNotFoundInAppliedOrderBenefitsException
        order_id,       # type: str
        benefit_id,     # type: str
    ):
        # type: (...) -> None

        self.order_id = order_id
        self.benefit_id = benefit_id

        super(Exception, self).__init__()

    def __str__(
        self,   # type: BenefitIdNotFoundInAppliedOrderBenefitsException
    ):
        # type: (...) -> str

        return "Benefit ID: {} not found in applied benefits for Order {}".format(self.benefit_id, self.order_id)


def _get_response_content(response):
    try:
        content = json.loads(response.content).get("detail").encode("utf8")
        return content
    except (AttributeError, ValueError):
        return ""


def build_message_exception(ex):
    if hasattr(ex, "response") and ex.response is not None:
        text = ex.response.text
        status_code = ex.response.status_code
        error_message = _get_error_message(
            status_code=status_code,
            text=text,
        )
        error_detail = {
            "type": "ClientError",
            "status_code": status_code,
            "message": error_message or None
        }
    else:
        error_type = type(ex).__name__
        error_message = ex.message
        error_detail = {"type": error_type, "message": error_message or None}

    ex.message = json.dumps(error_detail)
    return ex


def _get_error_message(
    status_code,    # type: int
    text,           # type: str
):
    # type: (...) -> str

    if status_code == 404:
        return "ERROR_COUPON_NOT_FOUND"
    if status_code == 406:
        return json.loads(text) if text not in (None, "None", "") else ""
    if status_code == 409:
        return "ERROR_COUPON_LOCKED_BY_ANOTHER_REQUEST"
    if status_code == 410:
        return "ERROR_COUPON_ALREADY_IN_USE"
    return ""



import json

from messagebus import (
    Event,
    AckMessage,
    Message,
    MessageBus,
    NakMessage,
)
from messageprocessorutil import UuidMessageProcessor
from typing import (
    Any,
    Optional,
)

from services.model import POSAlertNotificationRequest
from services.interactor import POSAlertNotificationInteractor


class POSAlertNotificationProcessor(UuidMessageProcessor):
    def __init__(
        self,
        pos_alert_notification_interactor,    # type: POSAlertNotificationInteractor
    ):
        # type: (...) -> None

        super(POSAlertNotificationProcessor, self).__init__()
        self.pos_alert_notification_interactor = pos_alert_notification_interactor

    def parse_data(
        self,
        data,           # type: str
        event_type,     # type: str
    ):
        # type: (...) -> POSAlertNotificationRequest

        request_json = json.loads(s=data)

        return POSAlertNotificationRequest(event_id=request_json.get('id'))

    def call_business(
        self,
        request,    # type: dict
    ):
        # type: (...) -> str

        return self.pos_alert_notification_interactor.process_request(request=request)

    def format_response(
        self,
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Optional[Event]
        request,        # type: Any
        result,         # type: Any
    ):
        # type: (...) -> None

        return self._do_not_reply_to_asynchronous_event()

    def format_parse_exception(
        self,
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Optional[Any]
        data,           # type: Any
        exception,      # type: Exception
    ):
        # type: (...) -> None

        return self._do_not_reply_to_asynchronous_event()

    def format_exception(
        self,
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Optional[Any]
        input_obj,      # type: POSAlertNotificationRequest
        exception,      # type: Exception
    ):
        # type: (...) -> None

        return self._do_not_reply_to_asynchronous_event()

    @staticmethod
    def _do_not_reply_to_asynchronous_event():
        return None



import os
import time
from threading import Lock

from typing import Optional

from systools import sys_parsefilepath

from application.model import UpdateType
from application.model.configuration import Configurations
from application.model.customexception import DisabledFunctionality
from application.repository.pos import BinaryUpdaterPOSRepository
from application.service import UpdaterService
from application.util import (
    LoggerUtil,
    settings
)

GENESIS_PATH = "../../genesis"

logger = LoggerUtil.get_logger_name(__name__)


class ChromeUpdaterService(object):
    def __init__(
        self,
        configs,            # type: Configurations
        working_on_update,  # type: Lock
        updater_service,    # type: UpdaterService
        pos_repository      # type: BinaryUpdaterPOSRepository
    ):
        # type: (...) -> None # noqa

        self.update_type = UpdateType.binaries
        self.update_name = self.update_type.name

        self.enabled = configs.updaters[self.update_name].enabled
        self.chrome_url = configs.updaters[self.update_name].chrome_url
        self.chrome_version_url = configs.updaters[self.update_name].chrome_version_url
        self.chrome_update_enabled = configs.updaters[self.update_name].chrome_update_enabled

        data_path = sys_parsefilepath("{HVDATADIR}")
        self.genesis_path = os.path.join(data_path, GENESIS_PATH)

        self.working_on_update = working_on_update
        self.updater_service = updater_service
        self.pos_repository = pos_repository

    def update_chrome(
        self,
        validate_version=False,     # type: bool
        restart_hv=False            # type: bool
    ):
        # type: (...) -> None

        if not self.chrome_update_enabled:
            return

        settings.set_update_control(self.update_name, None)
        if not self.enabled:
            settings.set_update_control(self.update_name, "$UPDATE_DISABLED")
            raise DisabledFunctionality()

        logger.info("Checking for Chrome")

        try:
            settings.set_update_control(self.update_name, "$UPDATING_CHROME")
            settings.set_update_control("updating", True)
            chrome_zip_path = self.genesis_path + '\\chrome.zip'
            chrome_version_file_exists = os.path.exists(os.path.join(self.genesis_path, "chrome", "chromeVersion.txt"))
            if not os.path.exists(self.genesis_path + '\\chrome') or not chrome_version_file_exists:
                validate_version = False

            if not validate_version:
                settings.set_update_control(self.update_name, "$GETTING_NEW_CHROME")
                logger.info("Downloading Chrome without validation")
                self._download_chrome(
                    file_path=chrome_zip_path,
                    restart_hv=restart_hv
                )
            else:
                logger.info("Downloading chromeVersion from server")
                settings.set_update_control(self.update_name, "$GETTING_CHROME_VERSION")
                chrome_version_server = self.pos_repository.download_from_url(
                    url=self.chrome_version_url
                )
                if chrome_version_server.status_code != 200:
                    logger.info("Error downloading Chrome")
                    raise

                with open(os.path.join(self.genesis_path, "chrome", "chromeVersion.txt"), 'r') as f:
                    chrome_version_installed = f.read().replace("\n", "")

                chrome_version_server = str(chrome_version_server.content)
                if chrome_version_server != chrome_version_installed:
                    settings.set_update_control(self.update_name, "$GETTING_NEW_CHROME")
                    logger.info("Downloading new Chrome version {}".format(chrome_version_server))
                    self._download_chrome(
                        file_path=chrome_zip_path,
                        new_chrome_version=chrome_version_server
                    )
                return
        except Exception as e:
            logger.error("Error trying to download Chrome {}".format(e))
            settings.set_update_control("updating", False)
            raise e
        settings.set_update_control("updating", False)

    def _download_chrome(
        self,
        file_path,               # type: str
        new_chrome_version="",   # type: str
        restart_hv=False         # type: bool
    ):
        # type: (...) -> None

        response = self.pos_repository.download_chrome_and_save(
            download_url=self.chrome_url,
            file_path=file_path
        )
        if response:
            path_to_extract = self.genesis_path + "\\chrome"
            self._do_backup(path=path_to_extract)

            try:
                settings.set_update_control(self.update_name, "$UNZIPPING_CHROME")
                self.updater_service.extract_zip(
                    zip_name=file_path,
                    path_to_extract=path_to_extract
                )
                self.updater_service.rename_folder(
                    original_path=path_to_extract + "\\chrome",
                    renamed_path=path_to_extract + "\\windows-x86"
                )
                if not new_chrome_version:
                    logger.info("Downloading Chrome version file")
                    new_chrome_version = self.pos_repository.download_from_url(url=self.chrome_version_url)
                    new_chrome_version = new_chrome_version.content if new_chrome_version.status_code == 200 else ""
                    with open(os.path.join(self.genesis_path, "chrome", "chromeVersion.txt"), 'w') as f:
                        f.write(new_chrome_version)
                self.updater_service.delete_directory(path=path_to_extract + "_backup")
                self.updater_service.delete_file(path=file_path)
            except (Exception,):
                logger.error("Error updating Chrome - Doing Rollback")
                self._do_backup(
                    path=path_to_extract,
                    restore=True
                )

            if restart_hv:
                logger.info("Starting Hypervisor to apply new Chrome")
                self.pos_repository.common_pos_repository.restart_hv(
                    event_type="POS_UPDATER_CHROME",
                )
                time.sleep(5)

            logger.info("Chrome update fully concluded with success!")
        else:
            logger.exception("Error downloading Chrome")

    def _do_backup(
        self,
        path,           # type: str
        restore=False   # type: Optional[bool]
    ):
        # type: (...) -> None

        from_path = path
        to_path = path + "_backup"

        if restore:
            from_path = path + "_backup"
            to_path = path
            self.updater_service.delete_directory(path=to_path)

        if not os.path.exists(from_path):
            return

        self.updater_service.rename_folder(
            original_path=from_path,
            renamed_path=to_path
        )



import datetime

from abc import ABCMeta

from typing import (
    Dict,
    Optional,
    List,
    Any,
)
from xml.etree import cElementTree as eTree

from _AposA8Align import AposA8Align
from _AposA8CommandType import AposA8CommandType
from _AposA8PrintLine import AposA8PrintLine
from _AposA8Helpers import (
    append_items_header,
    append_and_count_sale_items,
    APOS_DASH,
    APOS_NEWLINE,
    append_coupon_fiscal_parameters,
    build_call_id,
    get_text_line,
    format_text,
    split_lines,
    append_lines_from_array,
)
from common import PrinterCommon
from printer import FiscalPrinter
from msgbus import MBEasyContext

from clichehelper import ClicheHelper

from _AposA8FontSize import AposA8FontSize
from _AposA8Constants import AposA8Constants
from printer_wrapper import insert_space
from sysactions import translate_message
from models import LoyaltyConfigurations
from services import (
    get_fee_details,
    get_order_fees,
    order_xml_has_valid_fee,
)

AMOUNT_TEMPLATE_SMALL = "%-36s   R$%7.2f"
AMOUNT_TEMPLATE_LARGE = "%-20s   R$%7.2f"


class AposA8FiscalPrinter(FiscalPrinter):
    __metaclass__ = ABCMeta

    def __init__(
        self,
        mb_context,                     # type: MBEasyContext
        order_ref_ids,                  # type: list
        cliche_helper,                  # type: ClicheHelper
        customer_text_field,            # type: str
        coupon_fiscal_parameters,       # type: Dict[str, bool]
        font_size_withdrawal_panel,     # type: str
        danfe_font_size_mobile,         # type: str
        order_ref_position,             # type: str
        loyalty_configurations,         # type: LoyaltyConfigurations
    ):
        # type: (...) -> None

        super(AposA8FiscalPrinter, self).__init__(
            mb_context=mb_context,
            order_ref_ids=order_ref_ids,
            cliche_helper=cliche_helper,
            coupon_fiscal_parameters=coupon_fiscal_parameters,
            order_ref_position=order_ref_position,
            loyalty_configurations=loyalty_configurations,
        )

        self.customer_text_field = customer_text_field
        self.font_size_withdrawal_panel = font_size_withdrawal_panel
        self.danfe_font_size_mobile = danfe_font_size_mobile

    def print_cancel(self, posid, data):
        # not implemented
        pass

    def _build_loyalty_report(
        self,                    # type: AposA8FiscalPrinter
        model,                   # type: eTree.Element
        qr_code_text,            # type: str
        protocol_date,           # type: datetime
    ):
        # type: (...) -> List

        report = [
            APOS_NEWLINE,
            AposA8PrintLine(
                align=AposA8Align.CENTER,
                text=qr_code_text,
                command_type=AposA8CommandType.PRINT_QR_CODE,
            ),
            APOS_NEWLINE,
        ]

        if self.loyalty_configurations.text_to_print_in_loyalty_coupon:
            split_lines(
                report=report,
                text=self.loyalty_configurations.text_to_print_in_loyalty_coupon.encode("utf-8"),
                font_size=self.font_size_withdrawal_panel,
            )
            report.append(APOS_NEWLINE)
            report.append(APOS_NEWLINE)

        if self.loyalty_configurations.store_name:
            report.append(
                AposA8PrintLine(
                    align=AposA8Align.CENTER,
                    text="{}: {}".format(
                        translate_message(
                            model=model,
                            message_key="LABEL_STORE",
                        ),
                        self.loyalty_configurations.store_name,
                    ),
                    command_type=AposA8CommandType.PRINT_TEXT,
                    font_size=self.font_size_withdrawal_panel,
                )
            )
            report.append(APOS_NEWLINE)

        report.append(
            AposA8PrintLine(
                align=AposA8Align.CENTER,
                text="{}: {}".format(
                    translate_message(
                        model=model,
                        message_key="ISSUED_AT",
                    ),
                    protocol_date.strftime("%d/%m/%Y %H:%M:%S"),
                ),
                command_type=AposA8CommandType.PRINT_TEXT,
                font_size=self.font_size_withdrawal_panel,
            )
        )
        report.append(APOS_NEWLINE)

        return report

    def _append_total(
        self,
        report,         # type: List[AposA8PrintLine]
        order,          # type: eTree
        tenders,        # type: List[Any]
        items_amount,   # type: int
    ):
        # type: (...) -> None

        is_small_font = self.danfe_font_size_mobile == AposA8FontSize.SMALL

        cols_items_amount = 3
        column_width = self._get_column_width(subtract=cols_items_amount)
        label = format_text(
            text="QTD. TOTAL DE ITENS",
            width=column_width,
        )

        items_amount = format_text(
            text=("%3d" % items_amount).strip(),
            width=3,
            align='>'
        )

        qty_total = "{}{}".format(label, items_amount)
        report.append(self._get_text_line(text=qty_total))

        amount_template = AMOUNT_TEMPLATE_SMALL if is_small_font else AMOUNT_TEMPLATE_LARGE

        if order_xml_has_valid_fee(order_xml=order):
            order_fees = get_order_fees(order_xml=order)
            for fee in order_fees:
                fee_details = get_fee_details(fee_code=fee.code)

                fee_amount_text = amount_template % (fee_details.product_name.upper(), fee.amount)
                report.append(self._get_text_line(text=fee_amount_text))

        total_amount_value = float(order.get("totalAmount")) + float(order.get("taxTotal")) + float(order.get("tip", 0))

        total_amount_text = amount_template % ("VALOR TOTAL R$", total_amount_value)
        report.append(self._get_text_line(text=total_amount_text))

        discount_amount_value = float(order.get("discountAmount"))
        discount = amount_template % ("DESCONTO", discount_amount_value)
        report.append(self._get_text_line(text=discount))

        total_payable = amount_template % ("VALOR A PAGAR", total_amount_value - discount_amount_value)
        report.append(self._get_text_line(text=total_payable))

        label_value = "Valor Pago"
        column_width = self._get_column_width(subtract=len(label_value))

        label = format_text(
            text="FORMA DE PAGAMENTO",
            width=column_width,
        )

        payment = "{}{}".format(label, label_value)

        report.append(self._get_text_line(text=payment))

        for tender in tenders:
            tender_descr = self.get_payment_description(self.mb_context, tender)
            tender_amount = tender.amount
            tender_amount = ("R$%7.2f" % tender_amount).encode("utf-8")
            column_width = self._get_column_width(subtract=len(tender_amount))

            label = format_text(
                text=tender_descr,
                width=column_width,
            )

            tender_text = "{}{}".format(label, tender_amount)

            report.append(self._get_text_line(text=tender_text))

        # Get "Change" (sixth value of tender tuple) from last tender in tenders
        change = tenders[len(tenders) - 1].change
        if change > 0:
            order_voided = order.get("state") == "VOIDED"
            factor = -1 if order_voided else 1

            change_text = amount_template % ("Troco", change * factor)
            report.append(self._get_text_line(text=change_text))

    def _append_customer_text_field(
        self,
        report,     # type: List[AposA8PrintLine]
    ):
        # type: (...) -> None

        if self.customer_text_field != "":
            for customer_text in self.customer_text_field.split('\\n'):
                report.append(self._get_text_line(text=customer_text.replace("\\n", "")))

            report.append(APOS_DASH)

    def _append_satisfaction_survey(
        self,
        report,                          # type: List[AposA8PrintLine]
        model,                           # type: eTree
        satisfaction_survey_code,        # type: str
    ):
        # type: (...) -> None

        if satisfaction_survey_code is not None and model is not None:
            report.append(APOS_DASH)
            ssc = satisfaction_survey_code
            satisfaction_survey_code = '-'.join(ssc[i:i + 3] for i in range(0, len(ssc), 3))
            report.append(self._get_text_line(text=satisfaction_survey_code))
            report.append(APOS_DASH)

    def _append_coupon_body(
        self,
        model,           # type: eTree.Element
        order,           # type: eTree.Element
        report,          # type: List[AposA8PrintLine]
        tenders,         # type: List[Any]
        printer_common,  # type: PrinterCommon
    ):
        # type: (...) -> None

        append_items_header(
            report=report,
            font_size=self.danfe_font_size_mobile,
        )

        report.append(APOS_DASH)

        items_amount = append_and_count_sale_items(
            report=report,
            order=order,
            printer_common=printer_common,
            font_size=self.danfe_font_size_mobile,
        )

        report.append(APOS_DASH)

        self._append_total(
            report=report,
            order=order,
            tenders=tenders,
            items_amount=items_amount,
        )

        report.append(APOS_DASH)

        if self.order_ref_position == "BETWEEN_FISCAL_DATA":
            build_call_id(
                model=model,
                report=report,
                order=order,
                font_size_withdrawal_panel=self.font_size_withdrawal_panel,
                printer_common=printer_common,
                order_ref_ids=self.order_ref_ids,
            )

        append_coupon_fiscal_parameters(
            coupon_fiscal_parameters=self.coupon_fiscal_parameters,
            report=report,
            order=order,
            font_size=self.danfe_font_size_mobile,
        )

        satisfaction_survey_code = self.get_satisfaction_survey_code(order=order)
        self._append_satisfaction_survey(
            report=report,
            model=model,
            satisfaction_survey_code=satisfaction_survey_code,
        )

    def _append_cliche(
        self,
        report,     # type: List[AposA8PrintLine]
    ):
        # type: (...) -> None

        for line in self.cliche_helper.get_cliche():
            split_lines(
                report=report,
                text=line,
                font_size=self.danfe_font_size_mobile,
            )

        report.append(APOS_DASH)

    def _append_danfe_key(
        self,
        report,     # type: List[AposA8PrintLine]
        key,        # type: str
    ):
        # type: (...) -> None

        sat_key = str(insert_space(key.strip(), 4))
        font_is_small = True if self.danfe_font_size_mobile == AposA8FontSize.SMALL else False

        lines = [sat_key[:39], sat_key[40:]] if font_is_small else [sat_key[:29], sat_key[30:]]
        append_lines_from_array(
            report=report,
            lines=lines,
            font_size=self.danfe_font_size_mobile,
        )

    def _get_text_line(
        self,
        text,           # type: str
        align=None,     # type: Optional[int]
    ):
        # type: (...) -> AposA8PrintLine

        align = align if align is not None else AposA8Align.CENTER

        return get_text_line(
            align=align,
            text=text,
            font_size=self.danfe_font_size_mobile,
        )

    def _get_column_width(
        self,
        subtract,   # type: int
    ):
        # type: (...) -> int

        is_small_font = self.danfe_font_size_mobile == AposA8FontSize.SMALL
        column_width_small = AposA8Constants.COLS - subtract
        column_width_large = AposA8Constants.COLS_FONT_LARGE - subtract
        return column_width_small if is_small_font else column_width_large



import base64
import logging
import os
import sqlite3
import shutil
import time
from datetime import datetime
from sqlite3 import Connection
from threading import Thread
from xml.etree import cElementTree as eTree

import msgbus
import persistence
from persistence import (
    AprDbdException,
    Driver,
)
from typing import List, Tuple, Optional, Union
from old_helper import OrderTaker

from services import OrderService
from models import ErrorType

logger = logging.getLogger("FiscalWrapper")


class FiscalDataRepository:
    def __init__(
        self,
        mbcontext,              # type: msgbus.MBEasyContext
        database_dir="",        # type: Optional[str]
        schema_sql_path="",     # type: Optional[str]
    ):
        # type: (...) -> FiscalDataRepository

        self.mbcontext = mbcontext
        self.conn = None  # type: Optional[persistence.Connection]
        self.order_service = OrderService()
        self.database_dir = database_dir
        self.schema_sql_path = schema_sql_path

    def __enter__(self):
        self.conn = Driver().open(self.mbcontext, service_name="FiscalPersistence")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.conn:
            self.conn.close()

    def start_save_fiscal_data_thread(self, pos_id, order, xml_request, numero_sat, sent_to_nfce, xml_response):
        self.set_fiscal_xml_custom_property(pos_id, xml_request, order.get("orderId", ""))
        save_fiscal_data_thread = Thread(target=self.save_fiscal_data(pos_id, order, xml_request, numero_sat, sent_to_nfce, xml_response))
        save_fiscal_data_thread.daemon = True
        save_fiscal_data_thread.start()

    def set_fiscal_xml_custom_property(
        self,               # type: FiscalDataRepository
        pos_id,             # type: str
        xml_request,        # type: str
        order_id='',        # type: str
        key='FISCAL_XML',   # type: str
        blk_notify=True,    # type: bool
    ):
        # type: (...) -> None

        self.order_service.set_custom_property(
            pos_id=pos_id,
            key=key,
            value=xml_request,
            order_id=order_id,
            blk_notify=blk_notify,
        )

    def notify_pump(
            self,
            pos_id='',                      # type: str
            order_id='',                    # type: str
            xml_format='CancelledOrder',    # type: str
            order_picture=None              # type: Union[None, str]
    ):
        # type: (...) -> None

        try:
            event_type = "ORDER_REPROCESS"
            if not order_picture:
                order_picture = OrderTaker().get_order_picture(pos_id, order_id)

            xml = "<{}>{}</{}>".format(xml_format, order_picture, xml_format)
            self.mbcontext.MB_EasyEvtSend(event_type, event_type, xml)
        except Exception as _:
            logger.exception("Error notifying bohbump on POS ID: {} for ORDER ID: {}".format(pos_id, order_id))

    def save_fiscal_data(self, posid, order, xml_request, numero_sat, sent_to_nfce, xml_response):
        retries = 0
        while retries < 3:
            try:
                xml = base64.b64decode(xml_request)
                parsed_xml = eTree.XML(xml)

                from nfce import NfceRequestBuilder
                numero_nota_element = parsed_xml.find("infCFe/ide/nCFe")
                if numero_nota_element is None:
                    # TODO: Obter o numero da nota via parametro e remover esta logica daqui
                    numero_nota_element = parsed_xml.find(
                        "{{{0}}}NFe/{{{0}}}infNFe/{{{0}}}ide/{{{0}}}nNF".format(
                            NfceRequestBuilder.NAMESPACE_NFE))
                    if numero_nota_element is None:
                        numero_nota_element = parsed_xml.find(
                            "{{{0}}}infNFe/{{{0}}}ide/{{{0}}}nNF".format(
                                NfceRequestBuilder.NAMESPACE_NFE))

                numero_nota = numero_nota_element.text

                data_nota_element = parsed_xml.find("infCFe/ide/dEmi")
                hora_nota_element = parsed_xml.find("infCFe/ide/hEmi")
                if data_nota_element is None:
                    dh_elment = parsed_xml.find(
                        "{{{0}}}NFe/{{{0}}}infNFe/{{{0}}}ide/{{{0}}}dhEmi".format(
                            NfceRequestBuilder.NAMESPACE_NFE))
                    if dh_elment is None:
                        dh_elment = parsed_xml.find(
                            "{{{0}}}infNFe/{{{0}}}ide/{{{0}}}dhEmi".format(
                                NfceRequestBuilder.NAMESPACE_NFE))

                    data_nota_str = dh_elment.text

                    data_emissao_date_utc = data_nota_str[:19]
                    data_nota_date = datetime.strptime(data_emissao_date_utc, "%Y-%m-%dT%H:%M:%S")
                else:
                    data_nota_str = data_nota_element.text
                    hora_nota_str = hora_nota_element.text
                    data_nota_date = datetime.strptime(data_nota_str + hora_nota_str, "%Y%m%d%H%M%S")

                data_nota_seconds = int(time.mktime(data_nota_date.timetuple()))

                params_fiscal = {
                    "pos_id": posid,
                    "order_id": order.get("orderId"),
                    "data_nota": data_nota_seconds,
                    "numero_nota": numero_nota,
                    "numero_sat": numero_sat,
                    "xml_request": xml_request,
                    "sent_to_nfce": sent_to_nfce,
                    "xml_response": xml_response
                }

                self.conn.pquery("fiscal_saveFiscalData", **params_fiscal)
                break
            except Exception as _:
                logger.exception("Error saving fiscal data on database. RetryNum: {}".format(retries))
                time.sleep(0.2)
                retries += 1
        else:
            logger.error(">>> Save fiscal data failed after maximum retries!")

    def set_order_picture(self, order_id, order_picture):
        # type: (str, str) -> None
        params = {
            "order_id": order_id,
            "order_picture": order_picture
        }
        self.conn.pquery("fiscal_setOrderPicture", **params)

    def set_order_canceled(self, order_id):
        # type: (str) -> None
        params = {
            "order_id": order_id,
        }
        self.conn.pquery("fiscal_setOrderCanceled", **params)

    def _create_tables_fiscals(
        self,
        connection,     # type: Connection
    ):
        # type: (...) -> bool

        try:
            with open(self.schema_sql_path, "r") as f:
                query = f.read()

            if query:
                connection.executescript(query)

            return True
        except Exception as ex:
            logger.error("Error filling new_fiscal_persistcomp.db: {}".format(ex.message))
            return False

    def _backup_current_fiscal(
        self,
    ):
        # type: (...) -> bool

        try:
            logger.info("Creating backup database for fiscal_persistcomp.db")

            new_name = os.path.join(self.database_dir, "new_fiscal_persistcomp.db")
            connection = sqlite3.connect(database=new_name)

            filled_tables = self._create_tables_fiscals(connection=connection)
            if not filled_tables:
                os.remove(path=new_name)
                return False

            old_name = os.path.join(self.database_dir, "fiscal_persistcomp.db")
            time_format = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
            backup_name = os.path.join(self.database_dir, "fiscal_persistcomp.db_automatic_bkp_{}".format(time_format))
            success = False
            try:
                shutil.move(
                    src=old_name,
                    dst=backup_name,
                )
                success = True
            except WindowsError as ex:
                logger.error(ex.message)
                if os.path.exists(path=backup_name):
                    success = True

            if not success:
                return False

            try:
                shutil.move(
                    src=new_name,
                    dst=old_name,
                )
            except WindowsError as ex:
                logger.error(ex.message)
                pass

            if os.path.exists(path=new_name):
                connection.close()
                os.remove(path=new_name)

            return True

        except Exception as ex:
            logger.error("Error at backup of fiscal_persistcomp.db: {}".format(ex.message))
            return False

    def get_nfce_orders_to_send(
        self,
        limit,              # type: int
        first_try=True,     # type: Optional[bool]
    ):
        # type: (...) -> List

        try:
            params_fiscal = {
                "limit": limit
            }
            orders = [
                Order(
                    order_id=x.get_entry(0),
                    pos_id=x.get_entry(1),
                    xml=x.get_entry(2),
                    num_nota=x.get_entry(3),
                    order_picture=x.get_entry(4),
                )
                for x in self.conn.pselect("fiscal_getNfceToSend", **params_fiscal)
            ]
            return orders
        except AprDbdException as ex:
            if ex.errorcode == ErrorType.MALFORMED_DATABASE_DISK_IMAGE and first_try:
                logger.error("Error processing procedure fiscal_getNfceToSend: {}.".format(ex.message))

                backup_created = self._backup_current_fiscal()
                if not backup_created:
                    raise ex

                return self.get_nfce_orders_to_send(
                    limit=limit,
                    first_try=False,
                )
            raise ex

    def get_count_nfce_orders_to_send(self):
        # type: () -> int
        return [int(x.get_entry(0)) for x in self.conn.pselect("fiscal_getNfceToSendCount")][0]

    def set_nfce_sent(self, order_id, status):
        # type: (str, int) -> None
        params_fiscal = {
            "status": status,
            "order_id": order_id
        }
        self.conn.pquery("fiscal_setNfceSent", **params_fiscal)

    def set_xml_response(self, order_id, xml_base64):
        # type: (str, str) -> None
        params_fiscal = {
            "order_id": order_id,
            "xml_response": xml_base64
        }
        self.conn.pquery("fiscal_setXmlResponse", **params_fiscal)

    def set_nfce_sent_with_xml(self, order_id, xml_base64, status):
        # type: (str, str, int) -> None
        params_fiscal = {
            "status": status,
            "xml_request": xml_base64,
            "order_id": order_id
        }
        self.conn.pquery("fiscal_setNfceSentXml", **params_fiscal)

    def set_nfce_sent_with_xml_bkoffice(self, order_id, xml_base64, status, bkoffice_status):
        # type: (str, str, int, int) -> None
        params_fiscal = {
            "status": status,
            "xml_request": xml_base64,
            "order_id": order_id,
            "sent_to_bkoffice": bkoffice_status
        }
        self.conn.pquery("fiscal_setNfceSentXmlBKOffice", **params_fiscal)

    def retry_fiscal_orders_with_exception(self, range_of_days):
        # type: (str) -> None
        params_fiscal = {
            "range_of_days": "-{} days".format(range_of_days),
        }
        self.conn.pquery("fiscal_retryFiscalOrdersWithException", **params_fiscal)

    def get_xml_request(self, order_id):
        # type: (str) -> str
        params_fiscal = {
            "order_id": order_id
        }
        cursor = self.conn.pselect("fiscal_getRequestXMLFiscalData", **params_fiscal)
        if cursor.rows() > 0:
            ret = [x.get_entry(0) for x in cursor][0]
            return ret

    def get_xml_response(self, order_id):
        # type: (str) -> str
        params_fiscal = {
            "order_id": order_id
        }
        cursor = self.conn.pselect("fiscal_getResponseXMLFiscalData", **params_fiscal)
        if cursor.rows() > 0:
            ret = [x.get_entry(0) for x in cursor][0]
            return ret

    def get_numero_sat_and_xml_request(self, order_id):
        # type: (str) -> Tuple
        params_fiscal = {
            "order_id": order_id
        }
        cursor = self.conn.pselect("fiscal_getSatNumberAndRequestXMLFiscalData", **params_fiscal)
        if cursor.rows() > 0:
            ret = [(x.get_entry(0), x.get_entry(1)) for x in cursor][0]
            return ret


class Order:
    def __init__(self, order_id, pos_id, xml, num_nota, order_picture):
        self.order_id = order_id
        self.pos_id = pos_id
        self.xml = xml
        self.num_nota = num_nota
        self.order_picture = order_picture


class Tender(object):
    def __init__(self):
        # type: () -> Tender
        self.tender_seq_id = ""
        self.type = ""
        self.amount = ""
        self.change = ""
        self.bandeira = ""
        self.auth_code = ""
        self.cnpj_auth = ""
        self.receipt_customer = ""
        self.receipt_merchant = ""



from event_manager.async_response.dto import AsyncResponseRequest
from model import EventQueue, EventProcessorListener
from timeutil import Clock


class AsyncResponseInteractor(object):
    def __init__(
        self,
        event_queue: EventQueue,
        event_ack_listener: EventProcessorListener,
        clock: Clock,
    ):
        self.event_queue = event_queue
        self.event_ack_listener = event_ack_listener
        self.clock = clock

    def process_async_response(self, request: AsyncResponseRequest):
        publish_detail = self.event_queue.get_by_async_target_subject_and_id(
            reply_subject=request.subject,
            event_data=request.data,
        )
        if publish_detail is None:
            return
        if publish_detail.is_finished:
            return
        publish_detail.acked_at = self.clock.utc_now()
        self.event_ack_listener.event_processed(publish_detail=publish_detail)



import sys
import inspect

PY2 = sys.version_info[0] == 2


def with_metaclass(meta, *bases):
    # Taken from flask/six.
    class metaclass(meta):
        def __new__(cls, name, this_bases, d):
            return meta(name, bases, d)
    return type.__new__(metaclass, 'temporary_class', (), {})


if PY2:
    text_type = unicode
    string_type = basestring
    from itertools import izip_longest as zip_longest

    def with_str_method(cls):
        """Class decorator that handles __str__ compat between py2 and py3."""
        # In python2, the __str__ should be __unicode__
        # and __str__ should return bytes.
        cls.__unicode__ = cls.__str__
        def __str__(self):
            return self.__unicode__().encode('utf-8')
        cls.__str__ = __str__
        return cls

    def with_repr_method(cls):
        """Class decorator that handle __repr__ with py2 and py3."""
        # This is almost the same thing as with_str_method *except*
        # it uses the unicode_escape encoding.  This also means we need to be
        # careful encoding the input multiple times, so we only encode
        # if we get a unicode type.
        original_repr_method = cls.__repr__
        def __repr__(self):
            original_repr = original_repr_method(self)
            if isinstance(original_repr, text_type):
                original_repr = original_repr.encode('unicode_escape')
            return original_repr
        cls.__repr__ = __repr__
        return cls

    def get_methods(cls):
        for name, method in inspect.getmembers(cls,
                                               predicate=inspect.ismethod):
            yield name, method

else:
    text_type = str
    string_type = str
    from itertools import zip_longest

    def with_str_method(cls):
        # In python3, we don't need to do anything, we return a str type.
        return cls

    def with_repr_method(cls):
        return cls

    def get_methods(cls):
        for name, method in inspect.getmembers(cls,
                                               predicate=inspect.isfunction):
            yield name, method



from dailygoals_services.model.repository import DailyGoalsRepository
from dailygoals_services.service import DailyGoalsService


class UpdateTotalSoldInteractor:
    def __init__(self, daily_goals_repository, daily_goals_service):
        # type: (DailyGoalsRepository, DailyGoalsService) -> None

        self.daily_goals_repository = daily_goals_repository
        self.daily_goals_service = daily_goals_service

    def update(self):
        # type: () -> None

        self.daily_goals_service.update_daily_goals()
